import re, warnings
import lookml.lib.language_data.config
import lookml.lib.language_data._allowed_children
class ws:
    #basic whitespace paramters
    s = '  '
    nl = '\n'

    #size of a list type object before it breaks onto multiple lines. Int for number of items, not string length
    list_multiline_threshold = 4
    dense_children_threshold = 2
    dense_str_len = 25

    #Regex Patterns
    mustachePattern = r'\$\{([a-z\._0-9]*)\}'
    conditionPattern = r'\{\%\s{1,3}condition\s([a-z\._0-9]*)\s\%\}'
    doubleBracesPattern = r'\{\{\s{0,10}([a-z\._0-9]*)\s{0,10}\}\}'
    parameterPattern = r'\{\%\s{1,3}parameter\s([a-z\._0-9]*)\s\%\}'
    filtersPattern = r' \_filters\[\s{0,10}\'([a-z\._0-9]*)\'\]'
    fullPattern = r'(' + mustachePattern + r'|' + conditionPattern + r'|' + parameterPattern + r'|' + doubleBracesPattern + r'|' + filtersPattern + r')'
    lookml_name = r'^(\+|[a-z_])[a-z_0-9]{1,150}$'
    view_pattern = r'^\s*view\:\s[a-z_]{1,150}\s\{(\s|.)*\}\s*$' #P3 pattern will match multiple views in a row
    explore_pattern = r'^\s*explore\:\s[a-z_]{1,150}\s\{(\s|.)*\}\s*$' #P3 pattern will match multiple views in a row
    join_pattern = r'^\s*join\:\s[a-z_]{1,150}\s\{(\s|.)*\}\s*$' #P3 pattern will match multiple views in a row
    field_pattern = r'^\s*(dimension|measure|filter|parameter|dimension_group)\:\s(\+|[a-z_])[a-z_0-9]{1,150}\s\{(\s|.)*\}\s*$'
    model_pattern = r'.*'
    manifest_pattern = r'.*'

def valid_name(name):
    return re.match(ws.lookml_name,name)

def possible_view_str(s):
    return re.match(ws.view_pattern,s)

def possible_field_str(s):
    return re.match(ws.field_pattern,s)

def parse_references(inputString):
    '''
    Uses regular expresssions to preduce an iterator of the lookml references in a string.
    result has the shape {'raw':'${exact.field_reference}','field':'exact.field_reference', fully_qualified_reference:True}
    '''
    for match in re.findall(ws.fullPattern, inputString):
    # for match in re.findall(r'(\$\{([a-z\._0-9]*)\}|\{\%\s{1,3}condition\s([a-z\._0-9]*)\s\%\}|\{\%\s{1,3}parameter\s([a-z\._0-9]*)\s\%\}|\{\{\s{0,10}([a-z\._0-9]*)\s{0,10}\}\}| \_filters\[\s{0,10}\'([a-z\._0-9]*)\'\])',inputString):
        #Collapse the results from findall
        result = ''.join(match[1:])
        #Replace the liquid value references
        if result.endswith('._value'):
            result = result.replace('._value','')
        #Check if a fully qualified reference was used
        fq = True if '.' in ''.join(match[1:]) else False

        yield {'raw':match[0],'field':result, 'fully_qualified_reference': fq }

class props:
    cfg = lookml.lib.language_data.config.config
    _allowed_children = lookml.lib.language_data._allowed_children._allowed_children
    keys_with_name = (
        'user_attribute_param',
        'param',
        'form_param',
        'option',
    )
    plural_keys = (
        'view',
        'measure',
        'dimension',
        'filter',
        'dimension_group',
        'access_filter',
        'bind_filter',
        'map_layer',
        'parameter',
        'set',
        'column',
        'derived_column',
        'include',
        'explore',
        'link',
        'when',
        'allowed_value',
        'named_value_format',
        'join',
        'datagroup',
        'access_grant',
        'sql_step',
        'action',
        'param',
        'form_param',
        'option',
        'user_attribute_param',
        'assert',
        'test',
        'aggregate_table',
        "application",
        "visualization",
        "override_constant",
        "constant",
        "local_dependency",
        "remote_dependency",
        "query"
    )
    field_types = (
         'dimension'
        ,'measure'
        ,'parameter'
        ,'filter'
        ,'dimension_group'
    )

#Custom Exception classes
class OperationError(Exception):
    '''
        Provides messaging for issues with the operand and type
    '''
    def __init__(self, obj, operand, obj2):
        self.message = 'operation not supported' + str(type(obj)) + ' ' + str(operand) + ' ' + str(type(obj2))
        super().__init__(self.message)

class DuplicatePrimaryKey(Exception):
    '''
        Provides messaging when a second primary key is attempted to be added
    '''
    def __init__(self, view, dim ):
        self.message = f'A primary key cannot be added because one already exists'
        super().__init__(self.message)

class CoexistanceError(Exception):
    '''
        Provides messaging when a second primary key is attempted to be added
    '''
    def __init__(self, existing, new ,additional_message=''):
        self.message = f'A {new._type()} property cannot be added when a(n) {existing._type()} already exists. {additional_message}'
        super().__init__(self.message)


# import lookml as lookml
# # import src.lookml.lang as lang
# # import src.lookml.project as Project
# import lookml.lkml as lkml
# from lookml import lookml,lkml
# from lookml.common import project
import lookml
import lookml.lkml as lkml
import unittest, copy, json
from pprint import pprint
import warnings
import configparser
config = configparser.ConfigParser()
import pprint, base64
config.read('lookml/tests/.conf/settings.ini')
#P2: subtraction hooks / warnings
#P3: put colin's recursive sql table finder in the library
#P2: test coverage for sets
#P3: test coverage for manifest files
#P2: organize the tests

class testAllProps(unittest.TestCase):
    def setUp(self):
        self.proj = lookml.Project(path='.tmp/testAllProps')
        self.test_model_file = self.proj.new_file('test_model.model.lkml')
        self.test_model_file.write()
    
    def type_for_value(self,t:dict):
        map = {
                 'sql':'public.order_items'
                ,'yesno':'yes'
                ,'string':'test_string'
                ,'string_unquoted':'test_string_unquoted'
                ,'include':''
                ,'options':''
                ,'list_unquoted':['a','b','c']
                ,'html':'<b>{{value}}</b>'
                ,'options_quoted':''
                ,'filters':[{'field1':'%val%'},{'field2':'>1'}]
                ,'expression':''
                ,'list_quoted':['x','y','z']
                ,'anonymous_construct':{}
                ,'named_construct':{'name':'foo'}
                ,'named_construct_single':{'name':'foo'}
                ,'anonymous_construct_plural':[{},{}]
                ,'sorts':''
            }
        return map[t['type']]

    def test_view(self):
        v = lookml.View('test')
        _allowed_children = lookml.lib.language_data._allowed_children._allowed_children
        config = lookml.lib.language_data.config.config

        for child in _allowed_children['view']:
            try:
                v.__setattr__(child,self.type_for_value(config[child]['view']))
            except:
                pass
        v + lookml.Dimension('foo')
        for child in _allowed_children['dimension']:
            try:
                v.foo.__setattr__(child,self.type_for_value(config[child]['dimension']))
            except:
                pass
        self.test_model_file + v
    
    def tearDown(self):
        self.test_model_file.write()
        

class testMain(unittest.TestCase):
    '''
        Read a view file, test its methods and manipulation
    '''
    def setUp(self):
        pass

    def test_model_file(self):
        pm = lkml.load(open('.tmp/pylookml/eav_example/eav.model.lkml'))
        p = lookml.Model(pm)
        print(p)

    def test_parse_print(self):
        self.parsed_view = lkml.load(open('lookml/tests/files/basic_parsing/basic.view.lkml'))
        self.myView = lookml.View(self.parsed_view['views'][0])
        print(self.myView)

    def test_filters(self):
        self.parsed_view = lkml.load(open('lookml/tests/files/basic_parsing/basic.view.lkml'))
        self.myView = lookml.View(self.parsed_view['views'][0])
        self.myView.sum_foo.filters.foo.contains('test')
        self.myView + '''
            measure: total_sales {type: sum sql: ${TABLE}.sales ;;}
        '''
        self.myView.total_sales + 'filters: [foo:">100"]'
        self.myView.total_sales.filters + {'bar':'>100'}
        self.assertEqual(self.myView.total_sales.filters.bar.value,'>100')
        self.assertEqual(self.myView.total_sales.filters.foo.value,'>100')
        self.assertEqual(self.myView.sum_foo.filters.foo.value,'%test%')

    def test_json_serialization(self):
        self.parsed_view = lkml.load(open('lookml/tests/files/basic_parsing/basic.view.lkml'))
        self.myView = lookml.View(self.parsed_view['views'][0])
        self.assertIsNotNone(self.myView._json())
        print(self.myView._json())

    def test_refs(self):
        self.parsed_view = lkml.load(open('lookml/tests/files/basic_parsing/basic.view.lkml'))
        self.myView = lookml.View(self.parsed_view['views'][0])
        for f in self.myView._fields():
            #full reference
            print('full ref __ref__: ',f.__ref__)
            #short reference
            print('short ref __refs__: ',f.__refs__)
            #full reference -- regex escaped
            print('full ref regex __refre__: ',f.__refre__)
            #Short reference -- regex escaped
            print('short ref regex __refsre__: ',f.__refsre__)
            #Raw Reference
            print('raw full ref __refr__: ',f.__refr__)
            #Raw refence short
            print('raw ref short __refrs__: ',f.__refrs__)
            #Raw Reference regex
            print('raw ref regex __refrre__: ',f.__refrre__)

    def test_addition(self):
        tmp = lookml.View('''
            view: foo {
                derived_table: {
                    explore_source: order_items {
                        column: id { field: id }
                        column: id_2 { field: id }
                    }
                }
            }
        ''')
        tmp.derived_table.explore_source.column + 'column: foo {}'
        tmp.derived_table.explore_source + 'limit: 500'
        self.assertTrue('column: id_2 ' in str(tmp.derived_table.explore_source))
        print(tmp)

    def test_tags(self):
        self.parsed_view = lkml.load(open('lookml/tests/files/basic_parsing/basic.view.lkml'))
        self.myView = lookml.View(self.parsed_view['views'][0])
        self.myView.transaction.tags + ['tag5','tag6','tag7','tag8']
        self.myView.transaction.tags - 'tag7'
        self.myView.transaction.tags - ['tag6','tag4']
        self.assertTrue('tag3' in self.myView.transaction.tags)
        for tag in self.myView.transaction.tags:
            if tag == 'tag8':
                self.myView.transaction.tags - tag
        self.assertCountEqual(self.myView.transaction.tags,['tag1','tag2','tag3','tag5'])

    def test_adhoc_model_test(self):
        x = '''
            connection: "my_data"
            view: foo {
                dimension: a {
                    type: string
                    primary_key: yes
                }
                dimension: b {
                    #primary_key: yes
                }
                dimension: c {}
            }
            view: bar {
                dimension: x {}
                dimension: y {}
                dimension: z {}
            }
            explore: bar {
                join: foo {}
                join: test12 {}
            }
            explore: foo {
                join: bar {}
                aggregate_table: blah {
                    query: {
                        dimensions: [dim1,dim2,dim3]
                        filters: [foo:"%myfoo%",bar:"mybar%"]
                        sorts: [dim1: asc]
                        pivots: [dim4]
                    }
                    materialization: {
                        datagroup_trigger: foo
                    }
                }
            }
        '''
        parsed = lkml.load(x)
        # print(parsed)
        m = lookml.Model(parsed)
        # print(str(m.__dict__))
        
        m.views.foo.a.primary_key.value = 'no'
        m.views.foo.b.primary_key = 'yes'
        print(str(m))
        # print(str(m.explores['foo']))
        # print(str(m.explores.foo))

    def test_file_model_binding(self):
        proj = lookml.Project(path='lookml/tests/files/basic_parsing')
        x = proj.file('basic.model.lkml')
        cool = ['test123','test456','test890']
        x.explores['trip'] + ''.join([f'''
         join: {item} {{}}
         ''' for item in cool])
        self.assertTrue('test890' in str(x.explores.trip))

    def test_other(self):
        self.parsed_view = lkml.load(open('lookml/tests/files/basic_parsing/basic.view.lkml'))
        self.myView = lookml.View(self.parsed_view['views'][0])
        # if isinstance(self.myView.sum_foo.sql,lookml.prop):
        # if not isinstance(self.myView.sum_foo.sql,tuple()):
        # if (True if self.myView.sum_foo.sql._type() == '' else False):
        #     print('wow')
        # pprint(self.myView.transaction.__dict__)
        # print(self.myView.foo.link[0])
        # if 'transaction' in self.myView:
        #     print(self.myView.transaction)
        #P1: list size changed during iteration unless call parentheses are added
        # for i in self.myView.bar():
        #     print(i)
        self.assertTrue(self.myView.suggestions)
        # print(type(self.myView.transaction.tags))
        # print(lang.props._allowed_children['dimension_group'])
        # print(self.myView.derived_table.sql.value)
        # for f in self.myView(type=lookml.Field):
        #     if f.type.value == 'string':
        #         self.myView + f'measure: sum_{f.name} {{ type: sum sql: {f.__refs__};; }}'
        # print(lang.props._allowed_children['dimension'])
        # self.myView.sum_foo.direction = 'row'
        # for field in self.myView(type=lookml.Field):
        #     if '${TABLE}' in field.sql:
        #         print(field.name,": ", field.sql)
        # for f in self.myView._first_order_fields():
        #     print(f.sql)
        # print(self.myView.__dict__)
        # print(self.myView.set.__dict__)
        # print(self.myView.set.set1)
        # self.myView.set.set1.fields + 'foo'
        print(self.myView.foo.link[0].url.value)
        # print(self.myView)
        # for p in self.myView.bar.__iter__(type=lookml.prop):
        # for p in self.myView.transaction(type=lookml.prop, sub_type='timeframes', exclude_subtype='timeframes'):
        #     # if isinstance(p,lookml.prop_list_unquoted):
        #     print(p)
    def test_plural_anonymous_constructs(self):
        raw = '''
        view: x {
            derived_table: {
                explore_source: order_items {
                    column: id { field: order_items.id }
                    column: date { field: order_items.created_date }
                    column: status { field: order_items.status }
                    bind_filters: {
                        from_field: created_date
                        to_field: date
                    }
                    bind_filters: {
                        from_field: order_items.id
                        to_field: id
                    }
                    bind_filters: {
                        from_field: order_items.status
                        to_field: status
                    }
                }
            }
                dimension: date { type: date }
                dimension: id { type: number }
                dimension: status { 
                    type: string 
                    link: {
                        url: "http://facebook.com"
                    }
                    link: {
                        url: "http://yahoo.com"
                    }
                    link: {
                        url: "http://myspace.com"
                    }
                    }
                measure: count { type: count }
        }
        '''
        x = lookml.View(raw)
        self.assertTrue(x.derived_table.explore_source.bind_filters[0].from_field.value=='created_date')
        i = 0
        for item in x.derived_table.explore_source.bind_filters:
            i +=1 
        self.assertEqual(i,3)
        x.status.link + 'link: { url: "http://foo.com" }'
        # print(x.status.link)
        
        self.assertTrue(x.status.link[3].url.value=='http://foo.com')
        x.status.link[3].remove()
        i=0
        #P3: make this a list over the generator and len()
        for link in x.status.link:
            i += 1
        self.assertEqual(i,3)

    def test_anonymous_construct(self):
        #P1 support addition like lookml objects / pull out common addition functions
        # add, insert, add_hook, sub, subhook etc
        pass

    def test_all_subscriptability(self):
        proj = lookml.Project(
            path='lookml/tests/files/pylookml_test_project'
        )
        print(proj['views/01_order_items.view.lkml']['views']['order_items']['id'].primary_key.value)
        # print(proj['order_items.view.lkml']['views']) 
        
        #test file from project
        #test view/explore/prop from file/model
        #test field/prop from view
        #test prop from field
        
    def test_constructor(self):
        a = lookml.View('''
            view: foo {
                sql_table_name: public.order_items ;;
                dimension: foo {}
                measure: count {}
            }
        ''')
        self.assertTrue(isinstance(a,lookml.View))
        self.assertEqual(a.name,'foo')
        b = lookml.View('foo')
        self.assertTrue(isinstance(b,lookml.View))
        self.assertEqual(b.name,'foo')
        c = lookml.View({'name':'foo'})
        self.assertTrue(isinstance(c,lookml.View))
        self.assertEqual(c.name,'foo')
        with self.assertRaises(Exception) as context:
            lookml.View('''
            view: foo {
                sql_table_name: public.order_items ;;
                dimension: foo {}
                measure: count {}
            }
            view: bar {}
            ''')
        self.assertTrue('contains more than one view' in str(context.exception))
        #P3 does not throw the right exception, something is wrong with the ws.view_pattern and checker function
        with self.assertRaises(Exception) as context:
            lookml.View('explore: foo {}')
        # self.assertTrue('contains more than one view' in str(context.exception))
        x = lookml.Dimension('foo')
        y = lookml.Measure('bar')
        z = lookml.Filter('baz')
        aa = lookml.Parameter('fizz')
        ab = lookml.Dimension_Group('buzz')
        print(x,y,z,aa,ab)
        x = lookml.Dimension('dimension: foo { type: string }')
        y = lookml.Measure('measure: bar { type: count }')
        z = lookml.Filter('filter: baz { type: date }')
        aa = lookml.Parameter('parameter: fizz { type: unquoted }')
        ab = lookml.Dimension_Group('dimension_group: buzz { type: time }')
        print(x,y,z,aa,ab)
        exp = lookml.Explore('my_explore')
        # jn = lookml.Join('my_join')
        # mdl = lookml.Model('my_model')
        # mnfst = lookml.Manifest('my_manifest')
        # print(exp,jn,mdl,mnfst)
        print(exp)
        exp1 = lookml.Explore('''explore: my_explore {
            join: foo {}
        }''')
        print(exp1)
        # mdl1 = lookml.Model('conection: "cool" ')
        # mnfst1 = lookml.Manifest('project_name: "example"')
        # print(exp1,jn1,mdl1,mnfst1)
        # print(exp1,jn1)
        x = lookml.Dimension('dimension: +mydim1234567590 {}')

    def test_sql_enahancement_methods(self):
        tmp = lookml.View(
            '''
            view: tmp {
                dimension: foo {
                    sql: ${TABLE}.foo ;;
                }
            }
            '''
        )
        #P3: add arguments and date casting 
        tmp.foo.sql.nvl()
        self.assertTrue('nvl' in tmp.foo.sql.value)
        print(tmp)

    def test_more_add_hooks(self):
        #P2: other rules that should be run: various coexistence errors
        #P3: run coexistence errors and warnings and put to CSV
        foo = lookml.View('view: foo {}')
        foo + 'dimension: foo {}'
        # throws an assertion error:
        # foo.foo + 'type: count'
    def test_adding_joins_to_explore(self):
        #P1 test adding joins to explores
        pass
    def test_adding_fields_to_view(self):
        #P1 test adding fields to a view
        pass

    def test_set_property(self):
        x = lookml.Dimension('''
        dimension: x {}
        ''')
        x.setProperty('sql','${TABLE}.foo')
        self.assertEqual(x.sql.value,'${TABLE}.foo')
        y = lookml.View('y')
        y.setProperty('sql_table_name', 'public.order_items')
        self.assertEqual(y.sql_table_name.value,'public.order_items')

    def test_children_ancestors(self):
        # field.children() -> direct chilren
        #field.children_all() -> all chilren multi generation
        #field.dependency_chain(view=self.parent) -> tuple chain
        #view.print_dependency_map(format='human' or 'csv') -> print out of dep chains in view
        #field.print_dependency_map() -> print out of dep chain at field level
        #field.ancestors() -> any directs in the field
        #field.ancestors_all() -> any fields referenced up to ${TABLE}
        #field.print_ancestor_map() -> prints all fields which are on ancestor chain
        v = lookml.View('''
            view: v {
                dimension: a {}
                dimension: c { sql: ${a} ;; }
                dimension: d { sql: ${c} ;; }
                dimension: e { sql: ${d}/${a} ;; }
                dimension: b { sql: ${a} ;; }
            }
        ''')
        # v.print_dependency_map()
        for i in v.e.ancestors():
            print(i.__refs__)

    def test_has_prop(self):
        x = lookml.Dimension('''
            dimension: x {
                description: "hello"
            }
        ''')
        self.assertTrue(x.hasProp('description'))
        self.assertFalse(x.hasProp('view_label'))

    def test_dependency_map(self):
        parsed_view = lkml.load(open('lookml/tests/files/the_look/views/01_order_items.view.lkml'))
        myView = lookml.View(parsed_view['views'][0])
        myView.print_dependency_map()
    
    def test_remove_fields(self):
        x = lookml.View('''
            view: x {
                dimension: a {}
                dimension: b { sql: IFNULL(${a},0) ;;}
                dimension: c { sql: ${a} ;; }
                dimension: x {}
                dimension: y {}
                dimension: z {}
                dimension: zz {}
            }
        ''')
        with self.assertWarns(UserWarning) as wrn:
            y = ['a','b','c']
            for i in y:
                x - i
            for w in wrn.warnings:
                self.assertTrue('a had dependencies:' in str(w))
        self.assertTrue('a' not in x)
        self.assertTrue('b' not in x)
        self.assertTrue('c' not in x)
        with self.assertWarns(UserWarning) as wrn:
            x - 'sql_table_name'
            for w in wrn.warnings:
                self.assertTrue('sql_table_name did not exist on x' in str(w))
        x + 'sql_table_name: order_items ;;'
        x - 'sql_table_name'
        x.removeField(x.x)
        x.removeField('y')
        x - x.z
        x - 'zz'
        self.assertTrue('x' not in x)
        self.assertTrue('y' not in x)
        self.assertTrue('z' not in x)
        self.assertTrue('zz' not in x)

    def test_set_name_safe(self):
        x = lookml.View('''
            view: x {
                dimension: a {}
                dimension: b { sql: NVL(${a}) ;;}
                dimension: c { sql: ${a} ;; }
            }
        ''')
        x.a.setName_safe('foo')
        x.foo.setName_replace_references('bar')
        self.assertTrue('bar' in x)
        self.assertTrue('${bar}' in x.b.sql.value)
        self.assertTrue('${bar}' in x.c.sql.value)

    #P3: message / comments
    # def test_messages_comments(self): pass

    def test_legacy_methods(self):
        x = lookml.View('''
            view: x {
                dimension: test456 {
                    type: number
                    tags: ["my_tag"]
                }
                dimension: test123 {
                    type: string
                }
                measure: a {}
                measure: b {}
                dimension: test789 {}
                filter: c {}
                filter: d {}
                parameter: z {}
                parameter: y {
                    tags: ["my_tag"]
                }
                dimension_group: created {}
            }
        ''')
        self.assertTrue('test456' in x)
        self.assertFalse('imnotthere' in x)
        #
        fieldNames = list()
        for i in x.fieldNames():
            fieldNames.append(i)
        result = ['test456','test123','test789','a','b','c','d','z','y','created']
        self.assertEqual(fieldNames,result)
        #
        testSort = list()
        for i in x.getFieldsSorted():
            testSort.append(i.name)
        result = ['c','created','d','test123','test456','test789','y','z','a','b']
        self.assertEqual(result, testSort)
        #
        filterList = list()
        for i in x.filters():
            filterList.append(i.name)
        result = ['c','d']
        self.assertEqual(result, filterList)
        #
        paramList = list()
        for i in x.parameters():
            paramList.append(i.name)
        result = ['z','y']
        self.assertEqual(result, paramList)
        #
        dimList = list()
        for i in x.dimensions():
            dimList.append(i.name)
        result = ['test456','test123','test789']
        self.assertEqual(result, dimList)
        #
        measList = list()
        for i in x.measures():
            measList.append(i.name)
        result = ['a','b']
        self.assertEqual(result, measList)
        #
        dimGroupList = list()
        for i in x.dimensionGroups():
            dimGroupList.append(i.name)
        result = ['created']
        self.assertEqual(result, dimGroupList)
        #
        byTagList = list()
        for i in x.getFieldsByTag('my_tag'):
            byTagList.append(i.name)
        result = ['test456','y']
        self.assertEqual(byTagList,result)
        #
        byTypeList = list()
        for i in x.getFieldsByType('string'):
            byTypeList.append(i.name)
        result = ['test123','test789','c','d','z','y']
        self.assertEqual(byTypeList,result)
        #
        x.addDimension('orderItemID')
        self.assertTrue('order_item_id' in x)
        testDim = lookml.Dimension('dimension: test_dim {}')
        x.addDimension(testDim)
        self.assertTrue('test_dim' in x)
        #
        x.addAverage(x.test456)
        self.assertTrue('test456_avg' in x)
        self.assertEqual(x.test456_avg.sql.value, '${x.test456}')
        #
        x.test123.setPrimaryKey()
        self.assertTrue(x._View__pk.name == 'test123')
        x.test123.setName('new')
        self.assertTrue(x._View__pk.name == 'new')
        x.new.unSetPrimaryKey()
        x.test456.setPrimaryKey()
        self.assertTrue(x._View__pk.name == 'test456')
        #
        x.new.setType('number')
        x.new.setDescription('this is my new dimension')
        new_view = lookml.View('''
        view: new_view {
            view_label: "hello"
        }
        ''')
        x.new.setViewLabel(new_view.view_label.value)
        self.assertEqual('hello',x.new.view_label.value)
        x.test456.setDBColumn( 'mycol', changeIdentifier=True)
        self.assertTrue('mycol' in x)
        self.assertEqual(x.mycol.sql.value,"${TABLE}.`mycol`")
        x.y.removeTag('my_tag')
        x.z.setString()
        x.a.addLink('http://yahoo.com{% condition foo %}{{value | uri_encode }}{% endcondition %}','Go to Yahoo')

    def test_refinements(self):
        #parse a file with multiple refinements on the same object
        #confirm they collapse to a single view file
        proj = lookml.Project(path='lookml/tests/files/basic_parsing')
        test_file = proj.file('refine.view.lkml')
        print(test_file)
        self.assertTrue('first' in test_file.views['+test1'])
        self.assertTrue('second' in test_file.views['+test1'])

    def test_filtered_measure(self):
        meas = lookml.Measure('total_money')
        meas.setProperty('group_label','foo')
        # meas.properties.addProperty('filters',{'field':'order_items.price','value':'>100'})
        # meas.setProperty('filters',[{'field':'order_items.price','value':'>100'}])
        meas +  ('filters: { field: order_items.price value:">' + str(5) + '" }')
        meas.setViewLabel('test_viewLabel')
        # print(filt)
        # meas + filt
        print(meas)

    def test_add_micro_units(self):
        testView = lookml.View('test_view')
        testView + 'dimension: id {}'
        testView + 'dimension: success {}'
        testView + '''
                derived_table: {
                    explore_source: order_items {
                    column: order_id {field: order_items.order_id_no_actions }
                    column: items_in_order { field: order_items.count }
                    column: order_amount { field: order_items.total_sale_price }
                    column: order_cost { field: inventory_items.total_cost }
                    column: user_id {field: order_items.user_id }
                    column: created_at {field: order_items.created_raw}
                    column: order_gross_margin {field: order_items.total_gross_margin}
                    derived_column: order_sequence_number {
                        sql: RANK() OVER (PARTITION BY user_id ORDER BY created_at) ;;
                    }
                    }
                    datagroup_trigger: ecommerce_etl
                }
        '''
        print(testView)

    def test_adding_property(self):
        v = lookml.View('test')
        v + '''
                derived_table: {
                    explore_source: order_items {
                    column: order_id {field: order_items.order_id_no_actions }
                    column: items_in_order { field: order_items.count }
                    column: order_amount { field: order_items.total_sale_price }
                    column: order_cost { field: inventory_items.total_cost }
                    column: user_id {field: order_items.user_id }
                    column: created_at {field: order_items.created_raw}
                    column: order_gross_margin {field: order_items.total_gross_margin}
                    derived_column: order_sequence_number {
                        sql: RANK() OVER (PARTITION BY user_id ORDER BY created_at) ;;
                    }
                    }
                    datagroup_trigger: ecommerce_etl
                }
        '''
        v + 'dimension: id {}'
        v.id + 'sql: ${TABLE}.id ;;'
        for item in ('a', 'b', 'c'):
            v + f'''
                dimension: {item}_id {{ 
                        sql: {v.id.__refs__} + {item} ;;
                    }}'''

            v + f'''measure: sum_of_{item} {{
                type: sum
                sql: ${{{item}_id}};;
            }}
            '''
        for f in v.measures():
            if f.type.value == 'sum':
                f.addTag('my function is to add') 
        
        ex = lookml.Explore(v.name)
        ex + '''join: test_2 {
                    from: test
                    type: left_outer
                    relationship: one_to_many
                    sql_on: ${testid} = ${test_2.id};;
                 }
        '''
        ex.join.test_2 + 'sql_on: foo ;;'
        # F = lookml.File(ex)
        # F + v
        print(v)
    

#P2: obtain the real list of timezones from Looker itself
#P3: add CLI support
#P3: option to omit defaults
#P3: add looker version numbers to the lang map and throw warning if prop depreicated or error if not yet supported

#P1: merge refinements that are present in a single file (currently only last)
# see lookml/tests/files/pylookml_test_project/models/queries_for_order_items.view.lkml

class testProjFile(unittest.TestCase):
    def setUp(self):
        pass
    
    def pylookml_test_add_file(self,proj):
        nf = proj.new_file('hello.view.lkml')
        nf + lookml.View('''
        view: hello {}
        ''')
        nf.write()

    def pylookml_test_delete_file(self,proj):
        f = proj.file('hello.view.lkml')
        f.delete()
        x = proj.new_file('hello2.view.lkml')
        x.write()
        proj.delete_file('hello2.view.lkml')

    def pylookml_test_mutate_file(self,proj):
        #P1: should be file not new file, this method 
        # should throw warning for nf'ing an existing
        nf = proj.file('scratch/subfolder/pylookml_scratch.view.lkml')
        if isinstance(proj,lookml.ProjectSSH):
            commitMessage = proj._commit_message
            t = f'SSH {proj._path}'
        elif isinstance(proj,lookml.ProjectGithub):
            commitMessage = proj._commit_message
            t = f'pyGithub {proj._path}'
        else:
            commitMessage = 'local filesystem'
            t = f'direct filesystem {proj._path}'
        nf.views.pylookml_scratch.test.description = f'{t} {commitMessage}'
        #write the file
        nf.write()
        nf.views.pylookml_scratch.test.sql = '${hello}'
        nf.write()
        #confirm write
        abc = proj.file('scratch/subfolder/pylookml_scratch.view.lkml')
        self.assertTrue('test' in abc.views.pylookml_scratch)

    def pylookml_test_project_routine(self,proj):
        # self.assertEqual(len(list(proj.view_files())),19)
        #access a file's deep object via [] syntax
        val_a = proj['views/01_order_items.view.lkml']['views']['order_items']['order_id'].action[0].url.value
        self.assertEqual(val_a,'https://hooks.zapier.com/hooks/catch/1662138/tvc3zj/')
        #access a file's deep object via .file() syntax
        val_b = proj.file('views/01_order_items.view.lkml').views.order_items.order_id.action[0].url.value
        self.assertEqual(val_b,'https://hooks.zapier.com/hooks/catch/1662138/tvc3zj/')
        #mutate a file
        self.pylookml_test_mutate_file(proj)
        self.pylookml_test_add_file(proj)
        #proj.dir_list()
        self.pylookml_test_delete_file(proj)
        #test optional pyyaml

    def test_project_from_local_path(self):
        #connect
        proj = lookml.Project(
            path='lookml/tests/files/pylookml_test_project'
            )
        # pprint.pprint(proj._index)
        self.pylookml_test_project_routine(proj)

    def test_project_from_github(self):
        proj = lookml.Project(
            repo= "pythonruss/pylookml_test_project",
            access_token=config['project1']['access_token'],
            index_whole=True
        )
        self.pylookml_test_project_routine(proj)

    def test_project_from_github_fast(self):
        proj = lookml.ProjectGithub(
            repo= "pythonruss/pylookml_test_project",
            access_token=config['project1']['access_token'],
            index_whole=False
        )
        self.pylookml_test_project_routine(proj)


    def test_github_fast_file_exists_on_write(self):
        proj = lookml.ProjectGithub(
            repo= "pythonruss/pylookml_test_project",
            access_token=config['project1']['access_token'],
            index_whole=False
        )
        x = proj.file('views/01_order_items.view.lkml')
        x.views.order_items + 'dimension: test_github_fast_file_exists_on_write {}'
        x.write()

    def test_sha_not_supplied(self):
        #P2 reconstruct this test, was adhoc and stopped when it passed
        proj = lookml.ProjectGithub(
            repo= "llooker/aes_demo_final",
            access_token=config['autotune']['access_token'],
            branch=config['autotune']['branch'],
            index_whole=False
        )
        if proj._exists(f'pylookml/bike_share_aggs.view.lkml'):
            # click.echo('passed exists check')
            f = proj.file(f'pylookml/bike_share_aggs.view.lkml')
            # click.echo(f.sha)
        else:
            f = proj.new_file(f'pylookml/bike_share_aggs.view.lkml')
        f.write()

    def test_project_from_ssh(self):
        #connect
        proj = lookml.ProjectSSH(
             git_url='git@github.com:pythonruss/pylookml_test_project.git'
            ,looker_project_name='pylookml_test_project'
            ,looker_host='https://dat.dev.looker.com/'
            )
        self.pylookml_test_project_routine(proj)
        proj._git.add()
        proj.commit()
        proj._git.pushRemote()
        proj.deploy()
    
    def test_orphan_file(self):
        x = lookml.File('lookml/tests/files/basic_parsing/basic.view.lkml')
        self.assertTrue(isinstance(x,lookml.File))




class testOtherFiles(unittest.TestCase):
# Objective / coverage
#     Read all file types from filesystem: 
#     model, 
#     view, 
#     other lkml files, 
#     manifest, 
#     dashboard, 
#     js, 
#     json, 
#     maplayer
#     Read Files with isolated types of special syntax:
#     filters: old and new syntax
#     materializations
#     extensions
#     refinements
    def setUp(self):
        pass

    def test_parsing_aggregate_tables(self):
        x = lookml.File('lookml/tests/files/basic_parsing/agg.model.lkml')
        # x = lkml.load(open('lookml/tests/files/basic_parsing/agg.model.lkml','r', encoding="utf-8"))
        # print(str(x))
        print(x.explores.foo.aggregate_table.bar)

    def test_model_file(self):
        self.model_file = lookml.File('lookml/tests/files/basic_parsing/basic.model.lkml')
        print(str(self.model_file))

    def test_view_refinement(self):
        x = lookml.File('lookml/tests/files/basic_parsing/refine.view.lkml')
        print(str(x))

    def test_other_lkml_file(self):
        pass

    def test_manifest_file(self):
        x = lookml.File('lookml/tests/files/basic_parsing/manifest.lkml')
        #works
        self.assertEqual(x.remote_dependency['ga360'].url.value,'https://github.com/llooker/google_ga360')
        self.assertEqual(x.remote_dependency.ga360.url.value,'https://github.com/llooker/google_ga360')
        # print(x.remote_dependency['ga360'])
        # print(x.remote_dependency.ga360)
        self.assertTrue('ga360' in x.remote_dependency)
        for remote in x.remote_dependency:
            print(remote.override_constant)
        
        # print(type(x.contents))
        #anon: local_dependency, visualization, 
        #named: constant, application, 
        #other props / project name etc 
        
        # print(str(x.contents))

    def test_dashboard_file(self):
        lookml.lib.project.LOOKML_DASHBOARDS = True
        proj = lookml.Project(path='lookml/tests/files/the_look')
        x = proj.file('dashboards/brand_lookup.dashboard.lookml')
        self.assertEqual(x.content[0]['dashboard'],'brand_lookup')
        #P1 fix writing, currently lookml dashboards should be considered read only
        x.content[0]['dashboard'] = 'foo'
        # x.write()

    def test_js_file(self):
        pass
    def test_json_file(self):
        pass

    def test_maplayer_topojson(self):
        pass

#P1: Looping through view.views poduces a [None]
# Explore source??

class testExceptions(unittest.TestCase):
    def setup(self): pass
    def test_invalid_lookml_attribute(self):
        with self.assertWarns(UserWarning) as wrn:
            testView = lookml.View('test')
            testView + '''
                dimension: foo {
                    xxx: "yyy"
                }
            '''
            # Verify
            for w in wrn.warnings:
                self.assertTrue('xxx skipped. xxx not a valid attribute' in str(w))

    def test_duplicate_primary_key(self):
        #test to ensure that primary key is added, then if another primary key is set it throws an error, unless
        #the original primary key was set to no prior to the operation
        lookml.OMIT_DEFAULTS = True
        x = lookml.View('''
                view: x {
                    dimension: x {
                        primary_key: yes
                    }
                    dimension: y {}
                }
            ''')
        with self.assertRaises(lookml.lib.lang.DuplicatePrimaryKey) as context:
            x.y.primary_key = 'yes'
        x.x.primary_key = 'no'
        # print(x)
        x.y.primary_key = 'yes'
        #P2: document that this is the way to print and __pk does not return the pk
        # print(x)
        # print(x._View__pk)
    
    def test_coexistance_error(self):
        #checks to ensure the add_hook / exception process is working
        x = lookml.View('foo')
        x.sql_table_name = 'public.foo'
        with self.assertRaises(lookml.lib.lang.CoexistanceError):
            x + 'derived_table: { sql: select * from foo ;; }'
        del x.sql_table_name
        x + 'derived_table: { sql: select * from foo ;; }'
        with self.assertRaises(lookml.lib.lang.CoexistanceError):
            x.sql_table_name = 'public.foo'




class testWalks(unittest.TestCase):
  def setUp(self):
      self.proj = lookml.Project(path='lookml/tests/files/basic_parsing')
      self.view = self.proj.file('basic.view.lkml')

  def test_walking(self):
    for explore in self.view.explores:
      assert explore.name in ['basic']
      assert type(explore.join) == lookml.core.prop_named_construct

  def test_walk_explore(self):
    explore = self.view.explores.basic
    assert type(explore.join.cool) == lookml.core.prop_named_construct_single

  def test_walk_join(self):
    join = self.view.explores.basic.join.cool
    assert type(join) == lookml.core.prop_named_construct_single

    assert type(join.relationship) == lookml.core.prop_options
    assert join.relationship.value == 'many_to_one'
    
    assert type(join.type) == lookml.core.prop_options
    assert join.type.value == 'left_outer'
    
    assert type(join.sql_on) == lookml.core.prop_sql
    assert join.sql_on.value == '${basic.cool_id} = ${cool.basic_id}'
    # assert join.from == lookml.prop.string_unquoted

  def test_walk_view(self):
    view = self.view.views.basic
    assert type(view.extends) == lookml.core.prop_list_unquoted
    assert view.extends.value == ['base']

    assert type(view.extension) == lookml.core.prop_options
    assert view.extension.value == 'required'

    assert type(view.final) == lookml.core.prop_yesno
    assert view.final.value == 'no'

    assert type(view.label) == lookml.core.prop_string
    assert view.label.value == 'basic'

    assert type(view.view_label) == lookml.core.prop_string
    assert view.view_label.value == 'basic'

    assert type(view.required_access_grants) == lookml.core.prop_list_unquoted
    assert view.required_access_grants.value == ['a', 'b', 'c']

    assert type(view.suggestions) == lookml.core.prop_yesno
    assert view.suggestions.value == 'yes'

    assert type(view.derived_table) == lookml.core.prop_anonymous_construct

    for dim in view._dims():
      assert dim.name in ['foo', 'bar']

    assert type(view.transaction) == lookml.Dimension_Group

    for param in view._params():
      assert param.name in ['myparam']

    for measure in view._measures():
      assert measure.name in ['sum_foo', 'sum_bar']

  def test_walk_measure(self):
    measure = self.view.views.basic.sum_foo
    assert type(measure) == lookml.Measure
    assert measure.name == 'sum_foo'
    
    assert type(measure.type) == lookml.core.prop_options
    assert measure.type.value == 'sum'

    assert type(measure.sql) == lookml.core.prop_sql
    assert measure.sql.value == '${foo}'

    assert type(measure.filters) == lookml.core.prop_filters

  def test_walk_filters(self):
    filter = self.view.views.basic.sum_foo.filters.foo
    filter2 = self.view.views.basic.sum_bar.filters.foo

    assert type(filter) == lookml.core.flt
    assert filter.value == '%cool%'

    assert type(filter2) == lookml.core.flt
    assert filter2.value == '%cool%'

  def test_walk_dimension_group(self):
    dg = self.view.views.basic.transaction

    assert type(dg) == lookml.Dimension_Group
    assert dg.name == 'transaction'

    assert type(dg.type) == lookml.core.prop_options
    assert dg.type.value == 'time'

    assert type(dg.tags) == lookml.core.prop_list_quoted
    assert dg.tags.value == ['tag1', 'tag2', 'tag3', 'tag4']

    assert type(dg.timeframes) == lookml.core.prop_list_unquoted
    assert dg.timeframes.value == ['raw', 'time', 'date', 'week', 'month',
                                  'quarter', 'year', 'week_of_year', 'month_num']

    assert type(dg.sql) == lookml.core.prop_sql
    assert dg.sql.value == '${TABLE}.transaction_timestamp'

  def test_walk_dimension(self):
    dimension = self.view.views.basic.foo
    assert type(dimension) == lookml.core.Dimension

    assert type(dimension.type) == lookml.core.prop_options
    assert dimension.type.value == 'string'

    assert type(dimension.style) == lookml.core.prop_options
    assert dimension.style.value == 'classic'

    assert type(dimension.sql) == lookml.core.prop_sql
    assert dimension.sql.value == '${TABLE}.foo'

    assert type(dimension.link) == lookml.core.prop_anonymous_construct_plural

class testMicroUnits(unittest.TestCase):

    def test_add_orphan_to_github(self):
        proj = lookml.Project(
                 repo= config['github']['repo']
                ,access_token=config['github']['access_token']
                ,looker_host="https://profservices.dev.looker.com/"
                ,looker_project_name="russ_sanbox"
        )
        if proj._exists('lookml/tests/files/basic_parsing/refine.view.lkml'):
            x = proj.file('lookml/tests/files/basic_parsing/refine.view.lkml')
            proj.delete(x)
        f = lookml.File('lookml/tests/files/basic_parsing/refine.view.lkml')
        proj.put(f)
        f.delete()

    # def test_join_back_an_ndt(self):
    #     v = lookml.View('order_items')
    #     v + f'''
    #         sql_table_name: public.order_items ;;
    #         dimension: id {
    #             primary_key: yes
    #         }
    #         dimension: state {}
    #         dimension: sale_price {}
    #         parameter: {dynamic_dim_selector} {
    #             type: unquoted
    #         #     suggestions: ["Brand","Category","Department"]
    #             allowed_value: {
    #             label: "Category"
    #             value: "Category"
    #             }
    #             allowed_value: {
    #             label: "Brand"
    #             value: "Brand"
    #             }
    #             allowed_value: {
    #             label: "Department"
    #             value: "Department"
    #             }
    #             allowed_value: {
    #             label: "State"
    #             value: "State"
    #             }
    #         }
    #         dimension: user_id {}
    #         dimension: inventory_item_id { 
    #             sql: ${TABLE}.inventory_item_id ;; 
    #         }
    #         dimension: new_dimension {
    #             type: string
    #             sql:
    #                 {% if order_items.dynamic_dim_selector._parameter_value == 'Brand' %} ${products.brand}
    #                 {% elsif order_items.dynamic_dim_selector._parameter_value == 'Category' %}  ${products.category}
    #                 {% elsif order_items.dynamic_dim_selector._parameter_value == 'Department' %} ${products.department}
    #                 {% elsif order_items.dynamic_dim_selector._parameter_value == 'State' %} ${users.state}
    #                 {% else %} 'N/A'
    #                 {% endif %}
    #             ;;
    #         }
    #         measure: total_sale_price {
    #             type: sum
    #             sql: ${sale_price} ;;
    #         }
    #     '''
    #     ex = lookml.Explore(v.name)
    #     agg = lookml.View('agg')
    #     agg + '''
    #             derived_table: {
    #                 explore_source: order_items {
    #                 column: new_dimension {field: order_items.new_dimension}
    #                 column: total_sale_price {field: order_items.total_sale_price}
    #                 derived_column: rank {
    #                     sql: ROW_NUMBER() OVER (ORDER BY total_sale_price DESC) ;;
    #                 }
    #                 # bind_all_filters: yes
    #                 bind_filters: {
    #                     from_field: order_items.{dynamic_dim_selector}
    #                     to_field: order_items.{dynamic_dim_selector}
    #                 }
    #                 # bind_filters: {
    #                 #     from_field: order_items.created_date
    #                 #     to_field: order_items.created_date
    #                 # }
    #                 }
    #             }
    #             dimension: new_dimension {
    #                 sql: ${TABLE}.new_dimension ;;
    #             }
    #             dimension: rank {
    #                 type: number
    #                 hidden: yes
    #             }

    #             filter: tail_threshold {
    #                 type: number
    #                 hidden: yes
    #             }

    #             dimension: stacked_rank {
    #                 type: string
    #                 sql:
    #                         CASE
    #                         WHEN ${rank} < 10 then '0' || ${rank} || ') '|| ${new_dimension}
    #                         ELSE ${rank} || ') ' || ${new_dimension}
    #                         END
    #                 ;;
    #             }

    #             dimension: ranked_brand_with_tail {
    #                 type: string
    #                 sql:
    #                     CASE WHEN {% condition tail_threshold %} ${rank} {% endcondition %} THEN ${stacked_rank}
    #                     ELSE 'x) Other'
    #                     END

    #                 ;;
    #             }

    #             dimension: total_sale_price {
    #                 value_format: "$#,##0.00"
    #                 type: number
    #             }
    #     '''

    #     ex + '''
    #         join: inventory_items {
    #             type: left_outer
    #             relationship: one_to_many
    #             sql_on: ${order_items.inventory_item_id} = ${inventory_items.id} ;;
    #         }
    #         join: products {
    #             type: left_outer
    #             sql_on: ${inventory_items.product_id} = ${products.id} ;;
    #             relationship: many_to_one
    #         }
    #           join: users {
    #             type: left_outer
    #             sql_on: ${order_items.user_id} = ${users.id} ;;
    #             relationship: many_to_one
    #                   }
    #           join: agg {
    #             type: left_outer
    #             relationship: many_to_one
    #             sql_on: ${order_items.new_dimension}  = ${agg.new_dimension};;
    #         }
    #     '''
    #     myModel = lookml.File(ex)
    #     myModel + v
    #     myModel + agg
    #     myModel.properties.addProperty('connection', 'snowlooker')
    #     myModel.properties.addProperty('include', 'views/*.lkml')
    #     myModel.name = 'core2.model.lkml'
    #     proj = lookml.Project(
    #              repo= 'russlooker/oi'
    #             ,access_token=config['github']['access_token']
    #             ,looker_host="https://profservices.dev.looker.com/"
    #             ,looker_project_name="test_pylookml"
    #     )
    #     myModel
    #     proj.put(myModel)
    #     proj.deploy()

    def test_one_line_access_github(self):
        print(   
        lookml.Project(**config['project1'])['order_items.view.lkml']['views']['order_items']['id'].primary_key.value
        # lookml.Project(**config['project1']).file('order_items.view.lkml').views.order_items.id.primary_key.value
        )

    def test_local_file(self):
        x = lookml.File('lookml/tests/files/kitchenSink/kitchenSink.model.lkml')
        for v in x.views:
            for f in v.measures():
                if f.type.value == 'sum' and not f.name.endswith('_total'):
                    f.name = f.name + '_total'
        #Optionally Change the location
        # x.setFolder('.tmp')
        #Write the file
        x.write()

    def test_model_file_creation(self):
        #initialize project
        #create new model file
        #put and check output
        pass

    def test_parse_references(self):
        results = list(lookml.lib.lang.parse_references('''
            ${test.one_1} - ${test.two}
            {% condition test.three %} ${four} {% endcondition %}
            {% parameter test.five %}
            {{ six }}
            {{seven}}
            {{test.eight}}
            {{ _filters['test.nine'] | url_encode}}
            {% _filters['ten10'] %}
        '''))
        self.assertEqual(results[0]['field'],'test.one_1')
        self.assertEqual(results[1]['field'],'test.two')
        self.assertEqual(results[2]['field'],'test.three')
        self.assertEqual(results[3]['field'],'four')
        self.assertEqual(results[4]['field'],'test.five')
        self.assertEqual(results[5]['field'],'six')
        self.assertEqual(results[6]['field'],'seven')
        self.assertEqual(results[7]['field'],'test.eight')
        self.assertEqual(results[8]['field'],'test.nine')
        self.assertEqual(results[9]['field'],'ten10')
        self.assertEqual(results[9]['fully_qualified_reference'],False)

    def test_field_deletion(self):
        self.proj = lookml.Project(
                #  repo= config['github']['repo']
                # ,access_token=config['github']['access_token']
                 git_url='git@github.com:llooker/russ_sandbox.git'
                ,looker_host="https://dat.dev.looker.com/"
                ,looker_project_name="pylookml"
        )
        viewFile = self.proj.file('01_order_items.view.lkml')
        view = viewFile['views']['order_items'] 
        for f in view.fields():
            if f.name not in ('id'):
                view - f
        print(view)
        # or alternatively remove the fields with a loop
        # for field in viewFile.fields():
            # if field.name in ('count','sales_price'):
        
        # viewFile.removeField(field)


class testWriting(unittest.TestCase):
    '''
        Objective / coverage
          Read all file types from filesystem: 
            model, 
            view, 
            other lkml files, 
            manifest, 
            dashboard, 
            js, 
            json, 
            maplayer
            Success Criteria:
                need coverage of all basic operations: mutating each sub property and asserting the effect
                novel convenience methods should be covered in test cases
                written object needs to be reparsible to pass
    '''
    def setUp(self):
        pass

    #basic objects
    def test_model_file(self):
        pass
    def test_view_file(self):
        pass
    def test_other_lkml_file(self):
        pass
    def test_manifest_file(self):
        pass
    def test_dashboard_file(self):
        pass
    def test_js_file(self):
        pass
    def test_json_file(self):
        pass
    def test_maplayer_topojson(self):
        pass

    #syntax constructs
    def test_refinements(self):
        pass
    def test_extensions(self):
        pass
    def test_dimension(self):
        pass
    def test_measure(self):
        pass

    def tearDown(self):
        pass

class testModel(unittest.TestCase):
  def setUp(self):
      self.model = lookml.File('lookml/tests/files/basic_parsing/basic.model.lkml')
      self.explore_names = ['trip', 'station_weather_forecast', 'station_forecasting']

  def test_walking(self):
    for explore in self.model.explores:
      assert explore.name in self.model.explores
      assert isinstance(explore.join, lookml.core.prop_named_construct)

  def test_walk_explore(self):
    explore = self.model.explores.trip
    assert type(explore.join.start_station) == lookml.core.prop_named_construct_single

    explore2 = self.model.explores.station_weather_forecast
    assert type(explore2.hidden) == lookml.core.prop_yesno
    assert explore2.hidden.value == 'yes'
    assert type(explore.view_name) == lookml.core.prop_string_unquoted
    assert explore2.view_name.value == 'weather_forecast'
    
    # assert explore2.from == lookml.prop.string_unquoted

  def test_walk_join(self):
    join = self.model.explores.trip.join.start_station
    assert type(join) == lookml.core.prop_named_construct_single
    assert type(join.relationship) == lookml.core.prop_options
    assert join.relationship.value == 'many_to_one'
    assert type(join.type) == lookml.core.prop_options
    assert join.type.value == 'left_outer'
    assert type(join.sql_on) == lookml.core.prop_sql
    assert join.sql_on.value == '${trip.from_station_id} = ${start_station.station_id}'
    # assert join.from == lookml.prop.string_unquoted


import click
import looker_sdk
import lookml
import os, re, sys, json
import configparser
import time
from lookml.lib.utils import url

def init_looker_sdk(
        url:str, 
        client_id:str, 
        client_secret:str, 
        version:str='4.0') -> looker_sdk.methods40.Looker40SDK:
    os.environ['LOOKERSDK_BASE_URL'] = url
    os.environ['LOOKERSDK_CLIENT_ID'] = client_id
    os.environ['LOOKERSDK_CLIENT_SECRET'] = client_secret
    os.environ['LOOKERSDK_VERIFY_SSL'] = 'true'
    os.environ['LOOKERSDK_TIMEOUT'] = '120'
    os.environ['LOOKERSDK_API_VERSION'] = '4.0'
    return looker_sdk.init40()

def routine(
         access_token
        ,looker_host
        ,api_client
        ,api_secret
        ,model_name
        ,branch
    ):
    looker_host=url(looker_host)
    sdk = init_looker_sdk(looker_host.with_port_if_exists(), api_client, api_secret)
    frequent_queries = sdk.run_inline_query(
                result_format='json',
                body={
                    "model":"system__activity",
                    "view": "history",
                    "fields": [
                        "query.model",
                        "query.view",
                        "query.formatted_fields",
                        "query.filters",
                        "query.slug",
                        "history.query_run_count",
                        "history.max_runtime",
                        "history.average_runtime"
                    ],
                    "filters": {
                        # "query.model": f"-NULL,-\"system__activity\",{model_name}"
                        "query.model": f"{model_name}"
                    },
                    "sorts": [
                        "history.query_run_count desc"
                    ],
                    "limit": "100"
            })
    frequent_queries = json.loads(frequent_queries)
    click.echo("working on branch: " + branch)
    #look up the model attributes
    model = sdk.lookml_model(model_name)
    #look up the project attributes
    looker_project = sdk.project(model.project_name)
    if looker_project.git_service_name == 'github': 
        git_search = re.search(
            r'git@github.com:([a-zA-Z0-9_]{1,100}\/[a-zA-Z_0-9]{1,100})\.git', 
            looker_project.git_remote_url)
        if git_search:
            repo = git_search.group(1)

    pylookml_project = lookml.Project(
        repo= repo
        ,access_token=access_token
        ,looker_host=looker_host.with_no_port()+'/'
        ,looker_project_name=looker_project.name
        ,branch=branch
    )
    if pylookml_project._exists(f'pylookml/{model_name}_aggs.view.lkml'):
        # click.echo('passed exists check')
        f = pylookml_project.file(f'pylookml/{model_name}_aggs.view.lkml')
        # click.echo(f.sha)
    else:
        f = pylookml_project.new_file(f'pylookml/{model_name}_aggs.view.lkml')
    field_type_index = dict()
    #P1: we can't assume the file path, what if in subfolder?
    pylookml_model = pylookml_project[model_name+'.model.lkml']
    pylookml_model + f'includes: "pylookml/{model_name}_aggs.view"'
    #P1: we can't assume all explores are in the model file
    # lookml_model_explore has the filesystem location {
    # "source_file": "dbs.model.lkml"
    # }
    #P1: doesn't handle being run twice well
    with click.progressbar(frequent_queries) as fq:
        for query in fq:
                do = 0
                if looker_project.name not in field_type_index.keys():
                    do = 1
                    field_type_index[looker_project.name] = {}
                if query["query.model"] not in field_type_index[looker_project.name].keys():
                    do = 1
                    field_type_index[looker_project.name][query["query.model"]] = {}
                if query["query.view"] not in field_type_index[looker_project.name][query["query.model"]].keys():
                    do = 1
                    field_type_index[looker_project.name][query["query.model"]][query["query.view"]] = {
                         'dimensions':[]
                        ,'measures':[]
                        }
                if do:
                    metadata = sdk.lookml_model_explore(query["query.model"], query["query.view"],fields='fields')
                    for dimension in metadata.fields.dimensions:
                        field_type_index[looker_project.name][query["query.model"]][query["query.view"]]['dimensions'].append(dimension.name)
                    for measure in metadata.fields.measures:
                        field_type_index[looker_project.name][query["query.model"]][query["query.view"]]['measures'].append(measure.name)
                try:
                    
                    #P1 we don't need the explore technically since we're writing elsewhere
                    # pylookml_explore = pylookml_model.explores[query["query.view"]]
                    # {query["query.slug"]} 
                    # pylookml_explore + f'''
                    # click.echo(f'{query["query.view"]}')
                    explore_source_file_path = sdk.lookml_model_explore(
                            lookml_model_name=query["query.model"],
                            explore_name=query["query.view"],
                            fields='source_file'
                            )
                    explore_source_file = pylookml_project.file(explore_source_file_path.source_file)
                    pylookml_source_explore = explore_source_file.explores[query["query.view"]]
                    if 'persist_for' in pylookml_source_explore:
                        materialization = pylookml_source_explore.persist_for
                    elif 'persist_with' in pylookml_source_explore:
                        materialization = 'datagroup_trigger: ' + pylookml_source_explore.persist_with.value
                    elif 'persist_for' in explore_source_file:
                        materialization = explore_source_file.persist_for
                    elif 'persist_with' in explore_source_file:
                        materialization = 'datagroup_trigger: ' + explore_source_file.persist_with.value
                    else:
                        materialization = 'persist_for: "24 hours"'

                    f + 'include: "/**/*.model"'
                    if f'+{query["query.view"]}' not in f.explores:
                        f + f'''explore: +{query["query.view"]} {{ }}'''
                    f.explores['+' + query["query.view"]] + f'''
                        aggregate_table: auto_pylookml_{query["query.slug"]} {{
                            query: {{
                                dimensions: []
                                measures: []
                                description: "{looker_host.with_no_port()}/x/{query["query.slug"]}"
                                filters: []
                                }}
                            materialization: {{
                                {materialization}
                                }}
                            }}
                        '''
                    fields = json.loads(query["query.formatted_fields"])
                    for field in fields:
                        if field in field_type_index[looker_project.name][query["query.model"]][query["query.view"]]["dimensions"]:
                            f.explores['+'+query["query.view"]].aggregate_table["auto_pylookml_"+query["query.slug"]].query.dimensions + field
                        if field in field_type_index[looker_project.name][query["query.model"]][query["query.view"]]["measures"]:
                            f.explores['+'+query["query.view"]].aggregate_table["auto_pylookml_"+query["query.slug"]].query.measures + field
                    try:
                        if query["query.filters"]:
                            filters = json.loads(query["query.filters"])
                            click.echo(query["query.filters"] + ' | ' + filters)
                    except:
                        pass
                        # click.echo(f'failure with {query["query.filters"]}')
                    #P1: aggregate tables can't take paramter fields. Need to ensure each filter is not a param
                    if filters:
                        # f.explores['+'+query["query.view"]].aggregate_table["auto_pylookml_"+query["query.slug"]].query.filters + {k:v[1:][:-1] for k,v in filters.items()}
                        f.explores['+'+query["query.view"]].aggregate_table["auto_pylookml_"+query["query.slug"]].query.filters + {k:v for k,v in filters.items()}
                    # pylookml_model.write()
                    # pylookml_project.put(pylookml_model)
                except:
                    e = sys.exc_info()[0]
                    click.echo(f'Unexpected error: {e}')
    f.write()

        # except:
        #     click.echo(f'error, review your inputs and try again {sys.exc_info()[0]}')


@click.command()
# @click.option('--repository',  prompt='repo i.e. llooker/mycompany',help='github repo containing the project')
@click.option('--access_token',  prompt='your github access token',help='github personal access token')
@click.option('--looker_host',  prompt='looker host https://demo.looker.com/',help='looker host to work upon')
@click.option('--api_client',  prompt='Client ID',help='Looker API client ID')
@click.option('--api_secret',  prompt='API Secret Key',help='Looker API Secret')
@click.option('--model_name',  prompt='Looker Model Name',help='model to evaluate, i.e. hr_analytics')
@click.option('--branch',  prompt='git branch',help='master by default, can be your dev branch')
def guided(
         access_token 
        ,looker_host
        ,api_client
        ,api_secret
        ,model_name
        ,branch
        ):
    routine(
         access_token
        ,looker_host
        ,api_client
        ,api_secret
        ,model_name
        ,branch
    )

@click.command()
@click.option('--config_path',type=str,prompt='path to .ini file',help='')
def useconf(config_path):
    '''
    Your config should look like this:
    [autotune]
    access_token = xxx
    looker_host = https://mycompany.looker.com:19999
    api_client = xxx
    api_secret = yyy
    model_name = bike_share
    branch = dev-john-doe-yddt
    '''
    config = configparser.ConfigParser()
    # config.read('lookml/tests/settings.ini')
    config.read(config_path)
    routine(**config['autotune'])

@click.group(invoke_without_command=True)
def autotune():
    if os.path.exists('autotune.ini'):
        config = configparser.ConfigParser()
        config.read('autotune.ini')
        routine(**config['autotune'])
    else:
        pass


autotune.add_command(useconf)
autotune.add_command(guided)


"""Serializes a Python dictionary into a LookML string."""

from copy import deepcopy
from typing import Any, Dict, Iterator, Sequence, Union

from lookml.lkml.keys import (
    EXPR_BLOCK_KEYS,
    KEYS_WITH_NAME_FIELDS,
    PLURAL_KEYS,
    QUOTED_LITERAL_KEYS,
)


class Serializer:
    """Serializes a Python dictionary into a LookML string.

    Review the grammar specified for the Parser class to understand how LookML
    is represented. The grammar details the differences between blocks, pairs, keys,
    and values.

    Attributes:
        parent_key: The name of the key at the previous level in a LookML block
        level: The number of indentations appropriate for the current position
        field_counter: The position of the current field when serializing
            iterable objects
        base_indent: Whitespace representing one tab
        indent: An indent of whitespace dynamically sized for the current position
        newline_indent: A newline plus an indent string

    """

    def __init__(self):
        """Initializes the Serializer."""
        self.parent_key: str = None
        self.level: int = 0
        self.field_counter: int = 0
        self.base_indent: str = " " * 2
        self.indent: str = ""
        self.newline_indent: str = "\n"

    def increase_level(self) -> None:
        """Increases the indent level of the current line by one tab."""
        self.field_counter = 0
        self.level += 1
        self.update_indent()

    def decrease_level(self) -> None:
        """Decreases the indent level of the current line by one tab."""
        self.field_counter = 0
        self.level -= 1
        self.update_indent()

    def update_indent(self) -> None:
        """Sets the indent string based on the current level."""
        self.indent = self.base_indent * self.level
        self.newline_indent = "\n" + self.indent

    def is_plural_key(self, key: str) -> bool:
        """Returns True if the key is a repeatable key.

        For example, `dimension` can be repeated, but `sql` cannot be.

        The key `allowed_value` is a special case and changes behavior depending on its
        parent key. If its parent key is `access_grant`, it is a list and cannot be
        repeated. Otherwise, it can be repeated. The key `filters` is another special
        key associated to the new filters syntax, and it also cannot be repeated.

        """
        if key.endswith("s"):
            singular_key = key.rstrip("s")
            return singular_key in PLURAL_KEYS and not (
                singular_key == "allowed_value"
                and self.parent_key.rstrip("s") == "access_grant"
            )
        else:
            return False

    def serialize(self, obj: dict) -> str:  # noqa: D202
        """Returns a LookML string serialized from a dictionary."""

        def chain_with_newline():
            for key, value in deepcopy(obj).items():
                yield from self.write_any(key, value)
                yield "\n"

        return "".join(chain_with_newline())

    def expand_list(self, key: str, values: Union[Sequence, Dict]) -> Iterator[str]:
        """Expands and serializes a list of values for a repeatable key.

        This method is exclusively used for sequences of values with a repeated key like
        `dimensions` or `views`, which need to be serialized sequentially with a newline
        in between.

        Args:
            key: A repeatable LookML field type (e.g. "views" or "dimension_groups")
            values: A sequence of objects to be serialized

        Returns:
            A generator of serialized string chunks

        """
        modified_key = (
            key.rstrip("s") if key not in ("filters", "bind_filters") else key
        )
        for i, value in enumerate(values):
            if i > 0:
                yield "\n"
            yield from self.write_any(modified_key, value)

    def write_any(
        self, key: str, value: Union[str, list, tuple, dict]
    ) -> Iterator[str]:
        """Dynamically serializes a Python object based on its type.

        Args:
            key: A LookML field type (e.g. "suggestions" or "hidden")
            value: A string, tuple, or list to serialize

        Raises:
            TypeError: If input value is not of a valid type

        Returns:
            A generator of serialized string chunks

        """
        if isinstance(value, str):
            yield from self.write_pair(key, value)
        elif isinstance(value, (list, tuple)) or (
            isinstance(value, (dict)) and (key == "filters" or key == "sorts")
        ):
            if self.is_plural_key(key):
                yield from self.expand_list(key, value)
            else:
                yield from self.write_set(key, value)
        elif isinstance(value, dict):
            if key in KEYS_WITH_NAME_FIELDS or "name" not in value.keys():
                name = None
            else:
                name = value.pop("name")
            yield from self.write_block(key, value, name)
        else:
            raise TypeError("Value must be a string, list, tuple, or dict.")

        self.field_counter += 1

    def write_block(
        self, key: str, fields: Dict[str, Any], name: str = None
    ) -> Iterator[str]:
        """Serializes a dictionary to a LookML block.

        Args:
            key: A LookML field type (e.g. "dimension")
            fields: A dictionary to serialize (e.g. {"sql": "${TABLE}.order_id"})
            name: An optional name of the block (e.g. "order_id")

        Returns:
            A generator of serialized string chunks

        """
        if self.field_counter > 0:
            yield "\n"

        yield from self.write_key(key)
        if name:
            yield f"{name} " + "{"
        else:
            yield "{"

        if fields:
            self.parent_key = key
            self.increase_level()
            yield "\n"
            for i, (key, value) in enumerate(fields.items()):
                if i > 0:
                    yield "\n"
                yield from self.write_any(key, value)
            self.decrease_level()
            yield self.newline_indent

        yield "}"

    def write_set(self, key: str, values: Union[Sequence[str], Dict]) -> Iterator[str]:
        """Serializes a sequence to a LookML block.

        Args:
            key: A LookML field type (e.g. "fields")
            value: A sequence to serialize (e.g. ["orders.order_id", "orders.item"])

        Returns:
            A generator of serialized string chunks

        """
        # `suggestions` is only quoted when it's a set, so override the default
        force_quote = True if key == "suggestions" else False

        yield from self.write_key(key)
        yield "["

        if values:
            if len(values) > 5:
                self.increase_level()
                yield self.newline_indent
                for i, value in enumerate(values):
                    if i > 0:
                        yield f",{self.newline_indent}"
                    if key == "filters" or key == "sorts":
                        yield from self.write_labeled_set(value, values[value])
                    else:
                        yield from self.write_value(key, value, force_quote)
                self.decrease_level()
                yield self.newline_indent
            else:
                for i, value in enumerate(values):
                    if i > 0:
                        yield f", "
                    if key == "filters" or key == "sorts":
                        yield from self.write_labeled_set(value, values[value])
                    else:
                        yield from self.write_value(key, value, force_quote)
        yield "]"

    def write_pair(self, key: str, value: str) -> Iterator[str]:
        """Serializes a key and value to a LookML pair.

        Args:
            key: A LookML field type (e.g. "hidden")
            value: The value string (e.g. "yes")

        Returns:
            A generator of serialized string chunks

        """
        yield from self.write_key(key)
        yield from self.write_value(key, value)

    def write_key(self, key: str) -> Iterator[str]:
        """Serializes a key to LookML.

        Args:
            key: A LookML field type (e.g. "sql")

        Returns:
            A generator of serialized string chunks

        """
        yield f"{self.indent}{key}: "

    def write_value(
        self, key: str, value: str, force_quote: bool = False
    ) -> Iterator[str]:
        """Serializes a value to LookML, quoting it required by the key or forced.

        Args:
            key: A LookML field type (e.g. "hidden")
            value: The value string (e.g. "yes")
            force_quote: True if value should always be quoted

        Returns:
            A generator of serialized string chunks

        """
        if force_quote or key in QUOTED_LITERAL_KEYS:
            yield '"'
            yield value
            yield '"'
        elif key in EXPR_BLOCK_KEYS:
            yield value
            yield " ;;"
        else:
            yield value

    def write_labeled_set(self, key: str, value) -> Iterator[str]:
        """Serializes a labeled set with string keys to LookML
        (e.g. [created_at: "7 Days", user.status: "-disabled"])
        
        Args:
            key: An array key (e.g. "created_at")
            value: The value string (e.g. "7 Days")
        
        Returns:
            A generator of serialized string chunks
        
        """

        yield key
        yield ": "
        if value != "asc" and value != "desc":
            yield '"'
            yield value
            yield '"'
        else:
            yield value



import click
import lookml
import pprint

@click.command()
def doctree():
    '''
    Lists the parameter tree of LookML and links to documentation
    '''
    _allowed_children = lookml.lib.language_data._allowed_children._allowed_children
    config = lookml.lib.language_data.config.config
    click.echo(pprint.pprint(_allowed_children))



import src.lookml.lookml as lookml
import src.lookml.project as project
import src.lookml.lkml as lkml

import unittest, copy, re, sys
from pprint import pprint
import configparser, json
import looker_sdk
from looker_sdk import models, methods, init31
config = configparser.ConfigParser()
config.read('tests/settings.ini') 

class testScenariosAggGen(unittest.TestCase):
    '''
        Objective: test end to end common use case scenarios 
    '''
    def setUp(self):
        self.sdk = looker_sdk.init40("tests/api.ini")

    def test_dimension_add(self):
        x = lookml.Explore('''
        explore: foo {
            aggregate_table: auto_pylookml_NBDM6pP {
                query: {       
                dimensions: []      
                measures: []      
                description: ""      
                filters: [  ]       
                limit: 5000
                }
                materialization: {       
                sql_trigger_value: select 1 ;;
                } }
                }
        ''')
        x.aggregate_table['auto_pylookml_NBDM6pP'].query.dimensions + 'foo'
        filt = {'mock_data_aa.feature_clean': '"Login_Login"'}
        filt = {k:v[1:][:-1] for k,v in filt.items()}
        x.aggregate_table['auto_pylookml_NBDM6pP'].query.filters + filt
        print(x)

    def test_generate_aggregate_table(self):
        frequent_queries = self.sdk.run_inline_query(result_format='json',body={
                    "model":"system__activity",
                    "view": "history",
                    "fields": [
                        "query.model",
                        "query.view",
                        "query.formatted_fields",
                        "query.filters",
                        "query.slug",
                        "history.query_run_count",
                        "history.max_runtime",
                        "history.average_runtime"
                    ],
                    "filters": {
                        "query.model": "-NULL,-\"system__activity\""
                    },
                    "sorts": [
                        "history.query_run_count desc"
                    ],
                    "limit": "10"
                })
        frequent_queries = json.loads(frequent_queries)
        field_type_index = dict()
        for query in frequent_queries:
            #look up the model attributes
            model = self.sdk.lookml_model(query["query.model"])
            #look up the project attributes
            looker_project = self.sdk.project(model.project_name)
            if looker_project.git_service_name == 'github' and looker_project.name == 'dbs':
                git_search = re.search(
                    r'git@github.com:([a-zA-Z0-9_]{1,100}\/[a-zA-Z_0-9]{1,100})\.git', 
                    looker_project.git_remote_url)
                if git_search:
                    repo = git_search.group(1)
                do = 0
                if looker_project.name not in field_type_index.keys():
                    do = 1
                    field_type_index[looker_project.name] = {}
                if query["query.model"] not in field_type_index[looker_project.name].keys():
                    do = 1
                    field_type_index[looker_project.name][query["query.model"]] = {}
                if query["query.view"] not in field_type_index[looker_project.name][query["query.model"]].keys():
                    do = 1
                    field_type_index[looker_project.name][query["query.model"]][query["query.view"]] = {
                         'dimensions':[]
                        ,'measures':[]
                        }
                if do:
                    metadata = self.sdk.lookml_model_explore(query["query.model"], query["query.view"],fields='fields')
                    for dimension in metadata.fields.dimensions:
                        field_type_index[looker_project.name][query["query.model"]][query["query.view"]]['dimensions'].append(dimension.name)
                    for measure in metadata.fields.measures:
                        field_type_index[looker_project.name][query["query.model"]][query["query.view"]]['measures'].append(measure.name)
                try:
                    pylookml_project = project.Project(
                         repo= repo
                        ,access_token=config['project1']['access_token']
                        ,looker_host='https://dat.dev.looker.com/'
                        ,looker_project_name=looker_project.name
                    )
                    #P1: we can't assume the file path, what if in subfolder?
                    pylookml_model = pylookml_project[query["query.model"]+'.model.lkml']
                    #P1: we can't assume all explores are in the model file
                    # lookml_model_explore has the filesystem location {
                    # "source_file": "dbs.model.lkml"
                    # }
                    pylookml_explore = pylookml_model.explores[query["query.view"]]
                    # {query["query.slug"]} 
                    pylookml_explore + f'''
                        aggregate_table: auto_pylookml_{query["query.slug"]} {{
                            query: {{
                                dimensions: []
                                measures: []
                                description: "https://dat.dev.looker.com/x/{query["query.slug"]}"
                                filters: []
                                limit: 5000
                                }}
                            materialization: {{
                                sql_trigger_value: select 1 ;;
                                }}
                        }}
                    '''
                    fields = json.loads(query["query.formatted_fields"])
                    for field in fields:
                        if field in field_type_index[looker_project.name][query["query.model"]][query["query.view"]]["dimensions"]:
                            pylookml_explore.aggregate_table["auto_pylookml_"+query["query.slug"]].query.dimensions + field
                        if field in field_type_index[looker_project.name][query["query.model"]][query["query.view"]]["measures"]:
                            pylookml_explore.aggregate_table["auto_pylookml_"+query["query.slug"]].query.measures + field
                    filters = json.loads(query["query.filters"])
                    #P1: aggregate tables can't take paramter fields. Need to ensure each filter is not a param
                    pylookml_explore.aggregate_table["auto_pylookml_"+query["query.slug"]].query.filters + {k:v[1:][:-1] for k,v in filters.items()}
                    pylookml_project.put(pylookml_model)
                    #P1: put a https://dat.dev.looker.com/x/<<slug>> in the description
                except:
                    print("Unexpected error:", sys.exc_info()[0])


    def tearDown(self):
        pass




import click
import lookml
@click.command()
@click.option('--repository',  prompt='repo i.e. llooker/mycompany',help='github repo')
@click.option('--access_token',  prompt='your github access token',help='github access token')
def directory(repository, access_token):
    """Simple File Listing Capability"""
    proj = lookml.ProjectGithub(
         index_whole=True
        ,repo=repository
        ,access_token=access_token
    )
    proj.dir_list()



import src.lookml.lookml as lookml
import src.lookml.lkml as lkml
import unittest, copy
from pprint import pprint
import configparser, json
# from looker_sdk import client, models, methods
from looker_sdk import models, methods, init31
config = configparser.ConfigParser()
config.read('settings.ini')

class testScenariosGenerateFromSchema(unittest.TestCase):
    '''
        Objective: test end to end common use case scenarios 
    '''
    def setUp(self):
        pass
    
    def test_create_view_from_info_schema(self):
        def column_to_dimension(col):
            if col['LOOKER_TYPE'] == 'time':
                tmpDim = lookml.DimensionGroup(
                        lookml.lookCase(col['COLUMN_NAME'])
                        )
            else:
                tmpDim = lookml.Dimension(lookml.lookCase(col['COLUMN_NAME']))
            tmpDim.setType(col['LOOKER_TYPE'])
            tmpDim.sql = "${TABLE}." + col['COLUMN_NAME']
            return tmpDim

        sdk = init31("tests/api.ini")
        sql = """
                SELECT 
                    t.TABLE_NAME
                    ,t.TABLE_SCHEMA
                    ,t.COLUMN_NAME
                    ,t.DATA_TYPE
                    , CASE 
                        WHEN t.DATA_TYPE IN ('TIMESTAMP_LTZ') THEN 'time'
                        WHEN t.DATA_TYPE IN ('FLOAT','NUMBER') THEN 'number'
                        ELSE 'string' END as "LOOKER_TYPE"
                FROM 
                    information_schema.COLUMNS as t
                WHERE
                    1=1
                    AND t.table_name = 'ORDER_ITEMS'
                    AND t.table_schema = 'PUBLIC'
                LIMIT 100
        """
        query_config = models.WriteSqlQueryCreate(sql=sql, connection_id="snowlooker")
        query = sdk.create_sql_query(query_config)
        response = sdk.run_sql_query(slug=query.slug, result_format="json")
        response_json = json.loads(response)
        order_items = lookml.View('order_items_3')
        order_items.sql_table_name = 'PUBLIC.ORDER_ITEMS'

        for column in response_json:
            order_items + column_to_dimension(column)

        order_items.sumAllNumDimensions()
        order_items.addCount()

        proj = lookml.Project(
                 repo= config['github']['repo']
                ,access_token=config['github']['access_token']
                ,looker_host="https://profservices.dev.looker.com/"
                ,looker_project_name="russ_sanbox"
        )
        myNewFile = lookml.File(order_items)
        proj.put(myNewFile)
        proj.deploy()

    def tearDown(self):
        pass


_allowed_children = \
{
        "model": [
                "access_grant",
                "case_sensitive",
                "connection",
                "datagroup",
                "explore",
                "fiscal_month_offset",
                "include",
                "label",
                "map_layer",
                "named_value_format",
                "persist_for",
                "persist_with",
                "test",
                "view",
                "week_start_day"
        ],
        "explore": [
                "case_sensitive",
                "label",
                "persist_for",
                "persist_with",
                "extends",
                "extension",
                "final",
                "required_access_grants",
                "sql_table_name",
                "view_label",
                "access_filter",
                "aggregate_table",
                "always_filter",
                "always_join",
                "cancel_grouping_fields",
                "conditionally_filter",
                "description",
                "group_label",
                "from",
                "fields",
                "hidden",
                "join",
                "sql_always_where",
                "sql_always_having",
                "tags",
                "view_name",
                "query"
        ],
        "dimension": [
                "case_sensitive",
                "label",
                "drill_fields",
                "required_access_grants",
                "suggestions",
                "view_label",
                "description",
                "group_label",
                "hidden",
                "tags",
                "action",
                "allow_fill",
                "alpha_sort",
                "bypass_suggest_restrictions",
                "can_filter",
                "case",
                "datatype",
                "end_location_field",
                "fanout_on",
                "full_suggestions",
                "group_item_label",
                "html",
                "label_from_parameter",
                "link",
                "map_layer_name",
                "order_by_field",
                "primary_key",
                "required_fields",
                "skip_drill_filter",
                "start_location_field",
                "suggest_persist_for",
                "style",
                "sql",
                "sql_end",
                "sql_start",
                "tiers",
                "sql_longitude",
                "sql_latitude",
                "string_datatype",
                "units",
                "value_format",
                "value_format_name",
                "alias",
                "convert_tz",
                "suggestable",
                "type",
                "suggest_dimension",
                "suggest_explore"
        ],
        "filter": [
                "case_sensitive",
                "label",
                "required_access_grants",
                "suggestions",
                "view_label",
                "description",
                "group_label",
                "hidden",
                "tags",
                "bypass_suggest_restrictions",
                "datatype",
                "full_suggestions",
                "group_item_label",
                "required_fields",
                "suggest_persist_for",
                "sql",
                "alias",
                "convert_tz",
                "suggestable",
                "type",
                "suggest_dimension",
                "suggest_explore",
                "default_value"
        ],
        "view": [
                "label",
                "dimension",
                "dimension_group",
                "drill_fields",
                "extends",
                "extension",
                "filter",
                "final",
                "measure",
                "parameter",
                "derived_table",
                "required_access_grants",
                "set",
                "sql_table_name",
                "suggestions",
                "view_label"
        ],
        "measure": [
                "label",
                "drill_fields",
                "required_access_grants",
                "view_label",
                "description",
                "group_label",
                "hidden",
                "tags",
                "action",
                "can_filter",
                "datatype",
                "fanout_on",
                "group_item_label",
                "html",
                "label_from_parameter",
                "link",
                "order_by_field",
                "required_fields",
                "sql",
                "value_format",
                "value_format_name",
                "alias",
                "convert_tz",
                "suggestable",
                "type",
                "suggest_dimension",
                "suggest_explore",
                "approximate",
                "approximate_threshold",
                "allow_approximate_optimization",
                "direction",
                "filters",
                "list_field",
                "percentile",
                "precision",
                "sql_distinct_key"
        ],
        "dimension_group": [
                "label",
                "drill_fields",
                "required_access_grants",
                "view_label",
                "description",
                "group_label",
                "hidden",
                "tags",
                "allow_fill",
                "bypass_suggest_restrictions",
                "can_filter",
                "datatype",
                "fanout_on",
                "full_suggestions",
                "group_item_label",
                "html",
                "order_by_field",
                "skip_drill_filter",
                "sql",
                "sql_end",
                "sql_start",
                "alias",
                "convert_tz",
                "suggestable",
                "type",
                "suggest_dimension",
                "suggest_explore",
                "intervals",
                "timeframes"
        ],
        "parameter": [
                "label",
                "required_access_grants",
                "suggestions",
                "view_label",
                "description",
                "group_label",
                "hidden",
                "tags",
                "bypass_suggest_restrictions",
                "full_suggestions",
                "group_item_label",
                "required_fields",
                "suggest_persist_for",
                "alias",
                "convert_tz",
                "suggestable",
                "type",
                "suggest_dimension",
                "suggest_explore",
                "default_value",
                "allowed_value"
        ],
        "query": [
                "label",
                "description",
                "group_label",
                "filters",
                "limit",
                "sorts",
                "timezone",
                "dimensions",
                "measures",
                "pivots"
        ],
        "application": [
                "label",
                "url",
                "file",
                "entitlements"
        ],
        "visualization": [
                "label",
                "url",
                "file",
                "id",
                "sri_hash",
                "dependencies"
        ],
        "action": [
                "label",
                "url",
                "icon_url",
                "form_url",
                "param",
                "form_param",
                "user_attribute_param"
        ],
        "form_param": [
                "label",
                "description",
                "type",
                "name",
                "option",
                "required",
                "default"
        ],
        "option": [
                "label",
                "name"
        ],
        "map_layer": [
                "label",
                "url",
                "file",
                "extents_json_url",
                "feature_key",
                "format",
                "max_zoom_level",
                "min_zoom_level",
                "projection",
                "property_key",
                "property_label_key"
        ],
        "datagroup": [
                "label",
                "description",
                "max_cache_age",
                "sql_trigger"
        ],
        "when": [
                "label",
                "sql"
        ],
        "link": [
                "label",
                "url",
                "icon_url"
        ],
        "allowed_value": [
                "label",
                "value"
        ],
        "derived_table": [
                "persist_for",
                "sql",
                "cluster_keys",
                "create_process",
                "datagroup_trigger",
                "distribution",
                "distribution_style",
                "explore_source",
                "indexes",
                "partition_keys",
                "sortkeys",
                "sql_create",
                "sql_trigger_value",
                "table_compression",
                "table_format",
                "publish_as_db_view"
        ],
        "materialization": [
                "persist_for",
                "datagroup_trigger",
                "sql_trigger_value"
        ],
        "manifest": [
                "local_dependency",
                "remote_dependency",
                "localization_settings",
                "project_name",
                "constant",
                "application",
                "visualization"
        ],
        "join": [
                "required_access_grants",
                "sql_table_name",
                "view_label",
                "from",
                "fields",
                "sql",
                "type",
                "foreign_key",
                "outer_only",
                "relationship",
                "required_joins",
                "sql_foreign_key",
                "sql_on",
                "sql_where"
        ],
        "set": [
                "fields"
        ],
        "derived_column": [
                "sql"
        ],
        "named_value_format": [
                "value_format",
                "strict_value_format"
        ],
        "explore_source": [
                "filters",
                "expression_custom_filter",
                "limit",
                "column",
                "derived_column",
                "bind_all_filters",
                "bind_filters",
                "sorts",
                "timezone"
        ],
        "always_filter": [
                "filters"
        ],
        "test": [
                "explore_source",
                "assert"
        ],
        "aggregate_table": [
                "query",
                "materialization"
        ],
        "remote_dependency": [
                "url",
                "ref",
                "override_constant"
        ],
        "entitlements": [
                "local_storage",
                "navigation",
                "new_window",
                "allow_forms",
                "allow_same_origin",
                "core_api_methods",
                "external_api_urls",
                "oauth2_urls",
                "global_user_attributes",
                "scoped_user_attributes",
                "new_window_external_urls",
                "use_form_submit",
                "use_embeds"
        ],
        "param": [
                "name",
                "value"
        ],
        "user_attribute_param": [
                "name",
                "user_attribute"
        ],
        "case": [
                "when",
                "else"
        ],
        "column": [
                "field"
        ],
        "access_filter": [
                "field",
                "user_attribute"
        ],
        "bind_filters": [
                "from_field",
                "to_field"
        ],
        "access_grant": [
                "allowed_values",
                "user_attribute"
        ],
        "create_process": [
                "sql_step"
        ],
        "override_constant": [
                "value"
        ],
        "constant": [
                "value",
                "export"
        ],
        "local_dependency": [
                "project"
        ],
        "localization_settings": [
                "default_locale",
                "localization_level"
        ]
}


"""Parses a sequence of tokenized LookML into a Python object."""

import logging
from typing import Optional, Sequence, Type, Union

import lookml.lkml.tokens as tokens
from lookml.lkml.keys import PLURAL_KEYS

# Delimiter character used during logging to show the depth of nesting
DELIMITER = ". "


class Parser:
    r"""Parses a sequence of tokenized LookML into a Python object.

    This parser is a recursive descent parser which uses the grammar listed below (in
    PEG format). Each grammar rule aligns with a corresponding method (e.g.
    parse_expression).

    Attributes:
        tokens: A sequence of tokens
        index: The position in the token sequence being evaluated
        progress: The farthest point of progress during parsing
        logger: Own logging Logger instance
        depth: The level of recursion into nested expressions
        log_debug: A flag indicating that debug messages should be logged. This flag
            exits to turn off logging flow entirely, which provides a small
            performance gain compared to parsing at a non-debug logging level.

    Grammar:
        expression ← (block / pair / list)*
        block ← key literal? "{" expression "}"
        pair ← key value
        list ← key "[" csv? "]"
        csv ← (literal / quoted_literal) ("," (literal / quoted_literal))* ","?
        value ← literal / quoted_literal / expression_block
        key ← literal ":"
        expression_block ← [^;]* ";;"
        quoted_literal ← '"' [^\"]+ '"'
        literal ← [0-9A-Za-z_]+

    """

    def __init__(self, stream: Sequence[tokens.Token]):
        """Initializes the Parser with a stream of tokens and sets the index.

        Args:
            stream: Lexed sequence of tokens to be parsed

        Raises:
            TypeError: If an object in the stream is not a valid token

        """
        for token in stream:
            if not isinstance(token, tokens.Token):
                raise TypeError(
                    f"Type {type(token)} for {token} is not a valid token type."
                )
        self.tokens = stream
        self.index: int = 0
        self.progress: int = 0
        self.logger: logging.Logger = logging.getLogger(__name__)
        self.depth: int = -1
        self.log_debug: bool = self.logger.isEnabledFor(logging.DEBUG)

    def jump_to_index(self, index: int):
        """Sets the parser index to a specified value."""
        self.index = index

    def backtrack_if_none(method):  # noqa: D202
        """Decorates parsing methods to backtrack to a previous position on failure.

        This method sets a marker at the current position before attempting to run a
        parsing method. If the parsing method fails and returns `None`, it resets the
        index to the marker.

        It also keeps track of the farthest index of progress in case all parsing
        methods fail and we need to return a SyntaxError to the user with a character
        number.

        Args:
            method (Callable): Method to be decorated for backtracking

        """

        def wrapper(self, *args, **kwargs):
            mark = self.index
            self.depth += 1
            result = method(self, *args, **kwargs)
            self.depth -= 1
            if result is None:
                self.progress = (
                    self.index if self.index > self.progress else self.progress
                )
                self.jump_to_index(mark)
            return result

        return wrapper

    def peek(self) -> tokens.Token:
        """Returns the token at the current index."""
        return self.tokens[self.index]

    def advance(self, length: int = 1):
        """Moves the index forward by n characters.

        Args:
            length: The number of positions forward to move the index.

        """
        self.index += length

    def consume(self) -> tokens.Token:
        """Returns the current index character and advances the index by 1 token."""
        self.advance()
        return self.tokens[self.index - 1]

    def consume_token_value(self) -> Optional[str]:
        """Returns the value of the current index token, advancing the index 1 token."""
        token = self.consume()
        return token.value

    def check(self, *token_types: Type[tokens.Token]) -> bool:
        """Compares the current index token type to specified token types.

        Args:
            *token_types: A variable number of token types to check against

        Raises:
            TypeError: If one or more of the token_types are not actual token types

        Returns:
            bool: True if the current token matches one of the token_types

        """
        if self.log_debug:
            self.logger.debug(
                "%sCheck %s == %s",
                (1 + self.depth) * DELIMITER,
                self.peek(),
                " or ".join(t.__name__ for t in token_types),
            )
        for token_type in token_types:
            if not issubclass(token_type, tokens.Token):
                raise TypeError(f"{token_type} is not a valid token type.")
        if type(self.peek()) in token_types:
            return True
        else:
            return False

    def parse(self) -> dict:
        """Main method of this class and a wrapper for the expression parser."""
        return self.parse_expression()

    def update_tree(self, target: dict, update: dict):
        """Add one dictionary to an existing dictionary, handling certain repeated keys.

        This method is primarily responsible for handling repeated keys in LookML like
        `dimension` or `set`, which can exist more than once in LookML but cannot be
        repeated in a Python dictionary.

        This method checks the list of valid repeated keys and combines the values of
        that key in `target` and/or `update` into a list and assigns a plural key (e.g.
        `dimensions` instead of `dimension`).

        Args:
            target: Existing dictionary of parsed LookML
            update: New dictionary to be added to target

        Raises:
            KeyError: If `update` has more than one key
            KeyError: If the key in `update` already exists and would overwrite existing

        Examples:
            >>> from pprint import pprint
            >>> parser = Parser((tokens.Token(1),))

            Updating the target with a non-existing, unique key.

            >>> target = {"name": "foo"}
            >>> update = {"sql_table_name": "foo.bar"}
            >>> parser.update_tree(target, update)
            >>> pprint(target)
            {'name': 'foo', 'sql_table_name': 'foo.bar'}

            Updating the target with a non-existing, repeatable key.

            >>> target = {"name": "foo"}
            >>> update = {"dimension": {"sql": "${TABLE}.foo", "name": "foo"}}
            >>> parser.update_tree(target, update)
            >>> pprint(target)
            {'dimensions': [{'name': 'foo', 'sql': '${TABLE}.foo'}], 'name': 'foo'}

            Updating the target with an existing, repeatable key.

            >>> target = {"name": "foo", "dimensions": [{"sql": "${TABLE}.foo", "name": "foo"}]}
            >>> update = {"dimension": {"sql": "${TABLE}.bar", "name": "bar"}}
            >>> parser.update_tree(target, update)
            >>> pprint(target)
            {'dimensions': [{'name': 'foo', 'sql': '${TABLE}.foo'},
                            {'name': 'bar', 'sql': '${TABLE}.bar'}],
             'name': 'foo'}

        """
        keys = tuple(update.keys())
        if len(keys) > 1:
            raise KeyError("Dictionary to update with cannot have multiple keys.")
        key = keys[0]
        stripped_key = key.rstrip("s")
        if stripped_key in PLURAL_KEYS:
            plural_key = stripped_key + "s"
            if plural_key in target.keys():
                target[plural_key].append(update[key])
            else:
                target[plural_key] = [update[key]]
        elif key in target.keys():
            if self.depth == 0:
                self.logger.warning(
                    'Multiple declarations of top-level key "%s" found. '
                    "Using the last-declared value.",
                    key,
                )
                target[key] = update[key]
            else:
                raise KeyError(
                    f'Key "{key}" already exists in tree '
                    "and would overwrite the existing value."
                )
        else:
            target[key] = update[key]

    @backtrack_if_none
    def parse_expression(self) -> dict:
        """Returns a parsed LookML dictionary from a sequence of tokens.

        Raises:
            SyntaxError: If unable to find a matching grammar rule for the stream

        Grammar:
            expression ← (block / pair / list)*

        """
        if self.log_debug:
            grammar = "[expression] = (block / pair / list)*"
            self.logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)
        expression: dict = {}
        if self.check(tokens.StreamStartToken):
            self.advance()
        while not self.check(tokens.StreamEndToken, tokens.BlockEndToken):
            block = self.parse_block()
            if block is not None:
                self.update_tree(expression, block)
                continue

            pair = self.parse_pair()
            if pair is not None:
                self.update_tree(expression, pair)
                continue

            list = self.parse_list()
            if list is not None:
                expression.update(list)
                continue

            token = self.tokens[self.progress]
            raise SyntaxError(
                f"Unable to find a matching expression for '{token.id}' "
                f"on line {token.line_number}"
            )

        if self.log_debug:
            self.logger.debug(
                "%sSuccessfully parsed expression.", self.depth * DELIMITER
            )
        return expression

    @backtrack_if_none
    def parse_block(self) -> Optional[dict]:
        """Returns a dictionary that represents a LookML block.

        Returns:
            A dictionary with the parsed block or None if the grammar doesn't match

        Grammar:
            block ← key literal? "{" expression "}"

        Examples:
            Input (before tokenizing into a stream):
            ------
            "dimension: dimension_name {
                label: "Dimension Label"
                sql: ${TABLE}.foo ;;
            }"

            Output (dictionary):
            -------
            {
                "name": "dimension_name",
                "label": "Dimension Label",
                "sql": "${TABLE}.foo"
            }

        """
        if self.log_debug:
            grammar = "[block] = key literal? '{' expression '}'"
            self.logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)

        key = self.parse_key()
        if key is None:
            return key

        #Commented out to revert to old behavior with refinement treated as a seperate object
        # literal: Optional[str] = ""
        # if self.check(tokens.RefinementToken):
        #     self.advance()
        #     literal = "+"

        if self.check(tokens.LiteralToken):
            literal = self.consume_token_value()
            #Commented out and switched back to line above to revert to old behavior with refinement treated as a seperate object
            # literal = "{}{}".format(literal, self.consume_token_value())
        else:
            literal = None

        if self.check(tokens.BlockStartToken):
            self.advance()
        else:
            return None

        expression = self.parse_expression()

        if self.check(tokens.BlockEndToken):
            self.advance()

            block = {key: expression}
            if literal:
                block[key]["name"] = literal

            if self.log_debug:
                self.logger.debug(
                    "%sSuccessfully parsed block.", self.depth * DELIMITER
                )
            return block
        else:
            return None

    @backtrack_if_none
    def parse_pair(self) -> Optional[dict]:
        """Returns a dictionary that represents a LookML key/value pair.

        Returns:
            A dictionary with the parsed block or None if the grammar doesn't match

        Grammar:
            pair ← key value

        Examples:
            Input (before tokenizing into a stream):
            ------
            label: "Foo"

            Output (dictionary):
            -------
            {"label": "Foo"}

        """
        if self.log_debug:
            grammar = "[pair] = key value"
            self.logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)

        key = self.parse_key()
        value = self.parse_value()
        if key is None or value is None:
            return None

        pair = {key: value}
        self.logger.debug(self.depth * DELIMITER + "Successfully parsed pair.")
        return pair

    @backtrack_if_none
    def parse_key(self) -> Optional[str]:
        """Returns a string that represents a literal key.

        Returns:
            A dictionary with the parsed block or None if the grammar doesn't match

        Grammar:
            key ← literal ":"

        Examples:
            Input (before tokenizing into a stream)::
            ------
            label:

            Output (string):
            -------
            "label"

        """
        if self.log_debug:
            grammar = "[key] = literal ':'"
            self.logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)
        if self.check(tokens.LiteralToken):
            key = self.consume_token_value()
        else:
            return None

        if self.check(tokens.ValueToken):
            self.advance()
        else:
            return None

        if self.log_debug:
            self.logger.debug("%sSuccessfully parsed key.", self.depth * DELIMITER)
        return key

    @backtrack_if_none
    def parse_value(self) -> Optional[str]:
        """Returns a string that represents a value.

        Returns:
            A string with the parsed value or None if the grammar doesn't match

        Grammar:
            value ← literal / quoted_literal / expression_block

        Examples:
            Input (before tokenizing into a stream):
            ------
            1) "Foo"
            2) "${TABLE}.foo ;;"

            Output (string):
            -------
            1) "Foo"
            2) "${TABLE}.foo"

        """
        if self.log_debug:
            grammar = "[value] = literal / quoted_literal / expression_block"
            self.logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)
        if self.check(tokens.QuotedLiteralToken, tokens.LiteralToken):
            value = self.consume_token_value()
            if self.log_debug:
                self.logger.debug(
                    "%sSuccessfully parsed value.", self.depth * DELIMITER
                )
            return value
        elif self.check(tokens.ExpressionBlockToken):
            value = self.consume_token_value()
            if self.check(tokens.ExpressionBlockEndToken):
                self.advance()
            else:
                return None
            if self.log_debug:
                self.logger.debug(
                    "%sSuccessfully parsed value.", self.depth * DELIMITER
                )
            return value
        else:
            return None

    @backtrack_if_none
    def parse_list(self) -> Optional[dict]:
        """Returns a dictionary that represents a LookML list.

        Returns:
            A dictionary with the parsed list or None if the grammar doesn't match

        Grammar:
            list ← key "[" csv? "]"

        Examples:
            Input (before tokenizing into a stream):
            ------
            "timeframes: [date, week]"

            Output (dictionary):
            -------
            {"timeframes": ["date", "week"]}

        """
        if self.log_debug:
            grammar = "[list] = key '[' csv? ']'"
            self.logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)

        key = self.parse_key()
        if key is None:
            return key

        if self.check(tokens.ListStartToken):
            self.advance()
        else:
            return None

        csv = self.parse_csv()
        csv = csv if csv else []

        if self.check(tokens.ListEndToken):
            self.advance()
            list = {key: csv}
            if self.log_debug:
                self.logger.debug(
                    "%sSuccessfully parsed a list.", self.depth * DELIMITER
                )
            return list
        else:
            return None

    @backtrack_if_none
    def parse_csv(self) -> Optional[Union[list, dict]]:
        """Returns a list that represents comma-separated LookML values.

        Returns:
            A list with the parsed values or None if the grammar doesn't match

        Grammar:
            csv ← (literal / quoted_literal) ("," (literal / quoted_literal))* ","?

        Examples:
            Input (before tokenizing into a stream):
            ------
            1) "[date, week]"
            2) "['foo', 'bar']"

            Output (list):
            -------
            1) ["date", "week"]
            2) ["foo", "bar"]

        """
        if self.log_debug:
            grammar = '[csv] = (literal / quoted_literal) ("," (literal / quoted_literal))* ","?'
            self.logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)
        values: dict = {}

        if self.check(tokens.LiteralToken, tokens.QuotedLiteralToken):
            values[len(values)] = self.consume_token_value()
        else:
            return None

        preserve_dict = False
        while not self.check(tokens.ListEndToken):
            key: Optional[str] = ""
            if self.check(tokens.CommaToken):
                self.advance()
            elif self.check(tokens.ValueToken):
                key = values.pop(len(values) - 1)
                preserve_dict = True
                self.advance()
            else:
                return None

            if self.check(tokens.LiteralToken, tokens.QuotedLiteralToken):
                if len(key):  # type: ignore
                    values[key] = self.consume_token_value()
                else:
                    values[len(values)] = self.consume_token_value()
            elif self.check(tokens.ListEndToken):
                break
            else:
                return None

        if not preserve_dict:
            values_list: list = []
            for k in values.values():
                values_list.append(k)
            values = values_list  # type: ignore

        if self.log_debug:
            self.logger.debug(
                "%sSuccessfully parsed comma-separated values.", self.depth * DELIMITER
            )
        return values



#from lookml import core as lookml
import lookml
import click
import os
import looker_sdk
import json, re, sys, pkg_resources
import pprint
from .cli.autotune.autotune import autotune as autotune
from .cli.doctree.doctree import doctree as doctree
from .cli.directory.directory import directory as directory

#P3 CLI show children report for a specific field
#P3 CLI change a dimension name safely
#P3 CLI create LookML from a well known artifact from another tool
#P3 CLI apply block
#P3 CLI auto-tune your Looker Instance (create aggregate tables for top queries)
# write logs to a file in cwd
#P3 CLI add sums & averages of all number dimensions 
@click.group()
def cli():
    pass

@click.command()
def version():
    '''Provides currently installed PyLookML Version'''
    click.echo(pkg_resources.get_distribution('lookml').version)


cli.add_command(doctree)
cli.add_command(autotune)
cli.add_command(directory)
cli.add_command(version)

if __name__ == '__main__':
    cli()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lookml',
 'lookml.cli',
 'lookml.cli.autotune',
 'lookml.cli.directory',
 'lookml.cli.doctree',
 'lookml.lib',
 'lookml.lib.language_data',
 'lookml.lkml',
 'lookml.tests',
 'lookml.tests.files.the_look..github.workflows.demo-dashboard-validation',
 'lookml.tests.scenarios']

package_data = \
{'': ['*'],
 'lookml.tests': ['files/basic_parsing/*',
                  'files/kitchenSink/*',
                  'files/order_items.view.lkml',
                  'files/order_items.view.lkml',
                  'files/order_items.view.lkml',
                  'files/pylookml_test_project/*',
                  'files/pylookml_test_project/dashboards/*',
                  'files/pylookml_test_project/models/*',
                  'files/pylookml_test_project/scratch/subfolder/*',
                  'files/pylookml_test_project/views/*',
                  'files/test.dashboard.lookml',
                  'files/test.dashboard.lookml',
                  'files/test.dashboard.lookml',
                  'files/test_ndt.view.lkml',
                  'files/test_ndt.view.lkml',
                  'files/test_ndt.view.lkml',
                  'files/the_look/.github/*',
                  'files/the_look/.github/workflows/*',
                  'files/the_look/cool/*',
                  'files/the_look/dashboards/*',
                  'files/the_look/models/*',
                  'files/the_look/test.view.lkml',
                  'files/the_look/views/*',
                  'files/thelook/*']}

install_requires = \
['PyGithub>=1.47,<2.0',
 'PyYAML>=5.3.1,<6.0.0',
 'click>=7.1.2,<8.0.0',
 'looker-sdk>=0.1.3-beta.20,<0.2.0']

entry_points = \
{'console_scripts': ['lookml = lookml.__main__:cli']}

setup_kwargs = {
    'name': 'lookml',
    'version': '3.0.3',
    'description': 'A batteries included API for automating your LookML',
    'long_description': None,
    'author': 'Russell Garner',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
